import argparse
import sys
import json
import os
import tempfile
import subprocess

from shapely.geometry import shape


def main():
    parser = argparse.ArgumentParser(
        description="Egress KWCOCO data to T&E baseline framework structure")

    parser.add_argument('kwcoco_dataset_path',
                        type=str,
                        help="Path to KWCOCO dataset to egress")
    parser.add_argument('region_path',
                        type=str,
                        help="Path to input T&E Baseline Framework Region "
                             "definition JSON")
    parser.add_argument('output_path',
                        type=str,
                        help="S3 path for output JSON")
    parser.add_argument("-o", "--outbucket",
                        type=str,
                        required=True,
                        help="S3 Output directory for STAC item / asset "
                             "egress")
    parser.add_argument("--aws_profile",
                        required=False,
                        type=str,
                        help="AWS Profile to use for AWS S3 CLI commands")
    parser.add_argument("-d", "--dryrun",
                        action='store_true',
                        default=False,
                        help="Run AWS CLI commands with --dryrun flag")
    parser.add_argument("-n", "--newline",
                        action='store_true',
                        default=False,
                        help="Output as simple newline separated STAC items")
    parser.add_argument('-s', '--show-progress',
                        action='store_true',
                        default=False,
                        help='Show progress for AWS CLI commands')

    baseline_framework_kwcoco_egress(**vars(parser.parse_args()))

    return 0


def _kwcoco_to_stac_item(item_id,
                         kwcoco_dataset_path,
                         region_path,
                         self_s3_outpath,
                         kwcoco_s3_outpath):
    with open(region_path) as f:
        region = json.load(f)

    region_features = []
    for feature in region.get('features', ()):
        if ('properties' in feature
           and feature['properties'].get('type') == 'region'):
            region_features.append(feature)

    # WARNING: Big assumption here that we only ever have a single
    # feature with type 'region' in the region file
    if len(region_features) != 1:
        raise RuntimeError("Expecting one and only one feature of type "
                           "'region' in region file")
    region_feature = region_features[0]

    region_geometry = region_feature['geometry']
    region_bbox = list(shape(region_geometry).bounds)

    return {'type': 'Feature',
            'stac_version': '1.0.0',
            'stac_extensions': [],
            'id': item_id,
            'geometry': region_geometry,
            'bbox': region_bbox,
            'properties': {},
            'links': [{'rel': 'self',
                       'href': self_s3_outpath,
                       'type': 'application/json'}],
            'assets': {
                'kwcoco': {
                    'href': kwcoco_s3_outpath,
                    'type': 'application/json',
                    'title': 'KWCOCO Manifest'}}}


def baseline_framework_kwcoco_egress(kwcoco_dataset_path,
                                     region_path,
                                     output_path,
                                     outbucket,
                                     aws_profile=None,
                                     dryrun=False,
                                     newline=False,
                                     show_progress=False):
    if aws_profile is not None:
        aws_base_command =\
            ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    if dryrun:
        aws_base_command.append('--dryrun')

    if not show_progress:
        aws_base_command.append('--only-show-errors')

    item_id, _ = os.path.splitext(os.path.basename(kwcoco_dataset_path))
    self_s3_outpath = os.path.join(
        outbucket, item_id, '{}.json'.format(item_id))
    kwcoco_s3_outpath = os.path.join(
        outbucket, os.path.basename(kwcoco_dataset_path))
    output_stac_item = _kwcoco_to_stac_item(item_id,
                                            kwcoco_dataset_path,
                                            region_path,
                                            self_s3_outpath,
                                            kwcoco_s3_outpath)

    with tempfile.NamedTemporaryFile() as temporary_file:
        with open(temporary_file.name, 'w') as f:
            print(json.dumps(output_stac_item, indent=2), file=f)

        subprocess.run([*aws_base_command,
                        temporary_file.name, self_s3_outpath],
                       check=True)

    # Recursive copy entire KWCOCO directory (assumed that all cropped
    # imagery referenced in KWCOCO dataset resides in this directory)
    subprocess.run([*aws_base_command, '--recursive',
                    os.path.dirname(kwcoco_dataset_path), outbucket],
                   check=True)

    # Only have the one KWCOCO dataset STAC item in this case
    output_stac_items = [output_stac_item]
    if newline:
        te_output = '\n'.join((json.dumps(item) for item in output_stac_items))
    else:
        te_output = {'raw_images': [],
                     'stac': {
                         'type': 'FeatureCollection',
                         'features': output_stac_items}}

    with tempfile.NamedTemporaryFile() as temporary_file:
        with open(temporary_file.name, 'w') as f:
            if newline:
                print(te_output, end='', file=f)
            else:
                print(json.dumps(te_output, indent=2), file=f)

        command = [*aws_base_command, temporary_file.name, output_path]

        subprocess.run(command, check=True)

    return te_output


if __name__ == "__main__":
    sys.exit(main())



"""
References:
    https://smartgitlab.com/TE/evaluations/-/wikis/Accenture-TA-1-Processing-Status
"""
import os
import ubelt as ub
from watch.stac.stac_search_builder import SENSOR_TO_DEFAULTS


def print_provider_debug_information():
    """
    Helper to debug STAC endpoints and data availability.

    Summarize information about our hard-coded registered endpoints and query
    information about what other endpoints might exist.

    CommandLine:
        source $HOME/code/watch/secrets/secrets
        xdoctest -m watch.stac._notebook print_provider_debug_information
    """
    from rich import print
    print('Printing debug information about known and discoverable providers')
    rows = []
    for stac_code, stac_info in SENSOR_TO_DEFAULTS.items():

        endpoint = stac_info['endpoint']

        for collection in stac_info['collections']:
            rows.append({
                'stac_code': stac_code,
                'endpoint': endpoint,
                'collection': collection,
            })

    import pandas as pd
    df = pd.DataFrame(rows)
    print('Registered endpoints / collections / codes')
    print(df.to_string())
    unique_endpoints = df['endpoint'].unique()

    collection_to_endpoint = ub.udict(ub.group_items(df['endpoint'], df['collection']))
    endpoint_to_collections = ub.udict(ub.group_items(df['collection'], df['endpoint']))
    collection_to_num_endpoints = collection_to_endpoint.map_values(len)
    num_endpoints_to_collections = collection_to_num_endpoints.invert(unique_vals=False)
    multiendpoint_collections = num_endpoints_to_collections - {1}
    single_endpoint_collections = num_endpoints_to_collections.get(1, set())
    num_multi_collections = sum(map(len, multiendpoint_collections.values()))
    print(f'There are {len(unique_endpoints)} unique endpoints')
    print(f'There are {len(single_endpoint_collections)} collection that are unique to an endpoint')
    print(f'There are {num_multi_collections} collections that exist in multiple endpoints')
    print('single_endpoint_collections = {}'.format(ub.urepr(single_endpoint_collections, nl=1)))
    print('multiendpoint_collections = {}'.format(ub.urepr(multiendpoint_collections, nl=1)))
    print('unique_endpoints = {}'.format(ub.urepr(unique_endpoints, nl=1)))

    smart_stac_header = {
        'x-api-key': os.environ['SMART_STAC_API_KEY']
    }

    found_endpoint_to_catalog = {}
    found_endpoint_to_collections = {}

    import pystac_client
    for endpoint in unique_endpoints:
        print(f'Query {endpoint=}')

        pystac_headers = {}
        if endpoint == 'https://api.smart-stac.com':
            pystac_headers.update(smart_stac_header)

        try:
            catalog = pystac_client.Client.open(endpoint, headers=pystac_headers)
            collections = list(catalog.get_collections())
        except Exception:
            print(f'Failed to query {endpoint=}')
        else:
            print(f'Found {len(collections)} collections')
            found_endpoint_to_catalog[endpoint] = catalog
            found_endpoint_to_collections[endpoint] = collections

    unregistered_rows = []
    for endpoint, collections in found_endpoint_to_collections.items():
        print(f'\nCollections in endpoint={endpoint}')
        known_collection_names = set(endpoint_to_collections[endpoint])
        found_collection_names = {c.id for c in collections}
        registered_names = known_collection_names & found_collection_names
        misregistered_names = known_collection_names - found_collection_names
        unregistered_names = found_collection_names - known_collection_names
        if registered_names:
            print('Valid registered collections = {}'.format(ub.urepr(registered_names, nl=1)))
        if misregistered_names:
            print(f'!!! Collections are registered that dont exist {misregistered_names=}')
        if unregistered_names:
            print('There are unregistered collections = {}'.format(ub.urepr(unregistered_names, nl=1)))
            for name in unregistered_names:
                unregistered_rows.append({
                    'stac_code': None,
                    'endpoint': endpoint,
                    'collection': name,
                })
    print('Unregistered Collections')
    unregistered_df = pd.DataFrame(unregistered_rows)
    print(unregistered_df.to_string())

    full_df = pd.concat([df, unregistered_df])
    new_rows = full_df.to_dict(orient='records')
    for row in ub.ProgIter(new_rows, verbose=3):
        collection_name = row['collection']
        endpoint = row['endpoint']
        if endpoint in found_endpoint_to_collections:
            name_to_col = {c.id: c for c in found_endpoint_to_collections[endpoint]}
            if collection_name in name_to_col:
                catalog = found_endpoint_to_catalog[endpoint]
                collection = name_to_col[collection_name]
                row['title'] = collection.title

                is_unregistered = (unregistered_df[['endpoint', 'collection']] == [endpoint, collection_name]).all(axis=1).sum()
                is_registered = (df[['endpoint', 'collection']] == [endpoint, collection_name]).all(axis=1).sum()
                if is_registered:
                    assert not is_unregistered
                is_bad = not is_unregistered and not is_registered
                assert not is_bad
                row['registered'] = bool(is_registered)

                result = catalog.search(
                    collections=[collection_name],
                    max_items=1
                )
                found = list(result.items())
                row['has_items'] = len(found)

                print(row)
                print(collection.summaries.lists)

    for row in new_rows:
        if 'smart-stac' in row['endpoint']:
            collection_name = row['collection']
            if collection_name.startswith('ta1-'):
                parts = collection_name.split('-')
                if parts[-1] in {'1', '2', '3'}:
                    processing = '-'.join(parts[-2:])
                else:
                    processing = parts[-1]
                row['processing'] = processing

    new_df = pd.DataFrame(new_rows)
    new_df = new_df.sort_values(['processing', 'endpoint', 'collection'])
    # new_df['has_items'] = new_df['has_items'].fillna(False)
    new_df.loc[new_df['has_items'] == 1, 'has_items'] = True
    new_df.loc[new_df['has_items'] == 0, 'has_items'] = False
    new_df.loc[new_df['registered'] == 1, 'registered'] = True
    new_df.loc[new_df['registered'] == 0, 'registered'] = False
    print(new_df.to_string())


def check_processed_regions():
    """
    Print out a table of how many images / region / collection there are.

    CommandLine:
        source $HOME/code/watch/secrets/secrets
        xdoctest -m watch.stac._notebook check_processed_regions
    """
    import json
    import pystac_client
    from datetime import datetime as datetime_cls
    import watch
    import pandas as pd
    from rich import print

    dvc_data_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')

    # MODIFY AS NEEDED
    headers = {
        'x-api-key': os.environ['SMART_STAC_API_KEY']
    }

    base = ((dvc_data_dpath / 'annotations') / 'drop6')
    # base = dvc_data_dpath / 'annotations'

    region_dpath = base / 'region_models'
    region_fpaths = list(region_dpath.glob('*.geojson'))
    # region_fpaths = list(region_dpath.glob('*_C*.geojson'))
    # region_fpaths = list(region_dpath.glob('*_R*.geojson'))
    # region_dpath = dvc_data_dpath / 'golden_regions/region_models'
    # region_fpaths = list(region_dpath.glob('*_S*.geojson'))

    provider = "https://api.smart-stac.com"
    catalog = pystac_client.Client.open(provider, headers=headers)

    all_collections = list(catalog.get_collections())

    collections_of_interest = ['planet-dove', 'ta1-pd-acc', 'ta1-pd-ara', 'ta1-pd-str']

    from watch.utils import util_pattern
    pat = util_pattern.Pattern.coerce('ta1-*-acc*').to_regex()
    collections_of_interest = [c.id for c in all_collections if pat.match(c.id)]

    collections_of_interest = [
        # 'ta1-s2-acc-2',
        # 'ta1-ls-acc-2',
        # 'ta1-wv-acc-2',
        # 'ta1-pd-acc-2',

        'ta1-s2-acc-3',
        'ta1-ls-acc-3',
        'ta1-wv-acc-3',
        'ta1-pd-acc-3',

        # 'ta1-s2-acc',
        # 'ta1-s2-acc-1',
        # 'ta1-ls-acc',
        # 'ta1-ls-acc-1',
        # 'ta1-wv-acc',
        # 'ta1-wv-acc-1',
    ]

    #     'ta1-pd-acc',
    #     'ta1-pd-acc-1',
    #     'ta1-mixedgsd-acc-1',
    #     'ta1-30m-acc-1'
    # ]
    from watch.utils import util_progress
    from watch.cli.ta1_stac_to_kwcoco import summarize_stac_item

    peryear_rows = []
    peritem_rows = []
    raw_stac_items = []

    mprog = util_progress.ProgressManager()
    jobs = ub.JobPool(mode='thread', max_workers=20)
    with mprog, jobs:
        # Check that planet items exist
        for collection in mprog.progiter(collections_of_interest, desc='Query collections'):
            # Check that planet items exist in our regions
            region_to_results = {}
            region_iter = mprog.progiter(region_fpaths, desc=f'Submit query regions for {str(collection)}')
            for region_fpath in region_iter:
                with open(region_fpath) as file:
                    region_data = json.load(file)
                region_row = [f for f in region_data['features'] if f['properties']['type'] == 'region'][0]
                region_id = region_row['properties']['region_id']
                geom = region_row['geometry']
                start = region_row['properties']['start_date']
                end = region_row['properties']['end_date']
                if end is None:
                    # end = datetime_cls.utcnow().date()
                    end = datetime_cls.now().date().isoformat()

                item_search = catalog.search(
                    collections=[collection],
                    datetime=(start, end),
                    intersects=geom,
                    max_items=1000,
                )
                job = jobs.submit(list, item_search.items())
                job.region_id = region_id
                job.collection = collection

        for job in mprog(jobs.as_completed(), total=len(jobs), desc='collect results'):
            region_id = job.region_id
            collection = job.collection
            results = job.result()
            region_to_results[region_id] = results

            year_to_results = ub.udict(ub.group_items(results, key=lambda r: r.get_datetime().year))

            for year, year_results in year_to_results.items():
                for r in year_results:
                    summary = summarize_stac_item(r)
                    summary['collection'] = collection
                    summary['region_id'] = region_id
                    summary['year'] = year
                    r.extra_fields['collection'] = collection
                    r.extra_fields['region_id'] = region_id
                    r.extra_fields['year'] = year
                    peritem_rows.append(summary)
                    raw_stac_items.append(r)

                year_dates = [r.get_datetime() for r in year_results]
                min_date = min(year_dates)
                max_date = max(year_dates)
                peryear_rows.append({
                    'collection': collection,
                    'region_id': region_id,
                    'num_results': len(year_results),
                    'year': year,
                    'min_date': min_date.isoformat(),
                    'max_date': max_date.isoformat(),
                    # **year_oo_num
                })

    for row in peryear_rows + peritem_rows:
        if row['collection'].endswith('acc-2'):
            row['processing'] = 'acc-2'
        elif row['collection'].endswith('acc-3'):
            row['processing'] = 'acc-3'
        elif row['collection'].endswith('acc-1'):
            row['processing'] = 'acc-1'
        elif row['collection'].endswith('acc'):
            row['processing'] = 'acc'

        if '-ls' in row['collection']:
            row['sensor'] = 'L8'
        elif '-s2' in row['collection']:
            row['sensor'] = 'S2'
        elif '-wv' in row['collection']:
            row['sensor'] = 'WV'
        elif '-pd' in row['collection']:
            row['sensor'] = 'PD'

    item_df = pd.DataFrame(peritem_rows)

    if 0:
        # Specific queries
        for item in raw_stac_items:
            summary = summarize_stac_item(item)
            b = summary['eo_bands']
            b = b if isinstance(b, str) else '|'.join(sorted(b))
            if b != 'pan':
                if summary['datetime'].year < 2016 and item.extra_fields['region_id'] == 'KR_R002':
                    # print(f'item.extra_fields={item.extra_fields}')
                    # print('summary = {}'.format(ub.urepr(summary, nl=1)))
                    print(item.id)

        sub = item_df[item_df.year < 2016]
        sub = sub[sub.sensor == 'worldview']
        sub.loc[:, 'eo_bands'] = [b if isinstance(b, str) else '|'.join(sorted(b)) for b in sub['eo_bands']]
        sub.loc[:, 'asset_names'] = [b if isinstance(b, str) else '|'.join(sorted(b)) for b in sub['asset_names']]
        for region_id, group in sub.groupby('region_id'):
            histo = group.value_counts(['region_id', 'sensor', 'year', 'eo_bands', 'eo_cloud_cover'])
            print(histo.to_string())

    def pandas_aggregate2(data, func, axis=0):
        assert isinstance(func, dict)
        agg_func = {}
        drop_cols = []
        for k, v in func.items():
            if isinstance(v, str) and v == 'drop':
                drop_cols.append(k)
            else:
                agg_func[k] = v
        df2 = data.drop(drop_cols, axis=1)
        out = df2.iloc[0:1].reset_index(drop=True).iloc[0].copy()
        agg_cols = df2.agg(agg_func, axis=axis)
        out.loc[agg_cols.index] = agg_cols
        return out

    # Aggregate over years
    agg_rows = []
    df = pd.DataFrame(peryear_rows)
    # print(df.to_string())

    for region_id, group in df.groupby(['region_id', 'collection']):
        agg = pandas_aggregate2(group, func={
            'num_results': 'sum',
            'min_date': 'min',
            'max_date': 'max',
            'year': 'drop',
        }, axis=0)
        agg_rows.append(agg)
    agg_df = pd.DataFrame(agg_rows)

    agg_df = agg_df.sort_values(['region_id', 'collection'])
    print(agg_df.to_string())

    # df = df.sort_values(['processing', 'sensor', 'collection'])
    # piv = df.pivot(['region_id'], ['processing', 'sensor', 'collection'], ['num_results'])

    agg_df = agg_df.sort_values(['sensor', 'processing', 'collection'])
    piv = agg_df.pivot(index=['region_id'], columns=['sensor', 'processing', 'collection'], values=['num_results'])
    # piv = piv.astype(bool)
    print(piv.to_string())

    df = df.sort_values(['sensor', 'processing', 'collection', 'year'])
    piv = df.pivot(index=['region_id'], columns=['sensor', 'processing', 'collection', 'year'], values=['num_results'])
    print(piv.to_string())

    for c, group in df.groupby('collection'):
        group = group.sort_values('year')
        piv = group.pivot(index=['region_id'], columns=['sensor', 'processing', 'collection', 'year'], values=['num_results'])
        print(piv.to_string())

    requested_regions = {p.name.split('.')[0] for p in region_fpaths}
    missing_regions = requested_regions - set(piv.index)
    print('No results for regions = {}'.format(ub.urepr(missing_regions, nl=0)))

    # print(df.to_string())
    # print(f'region_id={region_id}')
    # print(f'results={results}')
    # print(f'collection={collection}')
    # print('region_to_results = {}'.format(ub.urepr(region_to_results, nl=1)))


def _devcheck_providers_exist():
    """
    develoepr logic to test to see if providers are working

    """
    # from watch.stac.stac_search_builder import _ACCENTURE_PHASE2_TA1_PRODUCTS
    # provider = _ACCENTURE_PHASE2_TA1_PRODUCTS['ta1-pd-acc']['endpoint']
    import pystac_client
    import os
    import ubelt as ub

    headers = {
        'x-api-key': os.environ['SMART_STAC_API_KEY']
    }
    provider = "https://api.smart-stac.com"
    catalog = pystac_client.Client.open(provider, headers=headers)
    print(ub.urepr(list(catalog.get_collections())))

    # item_search = catalog.search(collections=["ta1-mixedgsd-acc-1"])
    # print(f'item_search={item_search}')

    item_search = catalog.search(collections=["ta1-s2-acc"])
    item_search = catalog.search(collections=["ta1-wv-acc"])
    if 1:
        item_iter = iter(item_search.items())
        # View cloud cover
        item = next(item_iter)
        ccs = []
        for item in item_iter:
            cc = item.to_dict()['properties']['eo:cloud_cover']
            ccs.append(cc)
        import kwplot
        sns = kwplot.autosns()
        import pandas as pd
        df = pd.DataFrame({'cc': ccs})
        sns.histplot(data=df, x='cc')

    item_search = catalog.search(collections=["ta1-pd-acc"])
    item_search = catalog.search(collections=["ta1-pd-ara"])
    item_search = catalog.search(collections=["ta1-pd-str"])
    print(f'item_search={item_search}')



#!/usr/bin/env python3
"""
See Old Script:
    ~/code/watch/scripts/run_pseudolive_consolidate.py
"""
import os
import subprocess
import ubelt as ub
import scriptconfig as scfg


class PseudoliveConsolidateConfig(scfg.DataConfig):
    """
    Run pseudolive consolidation script for TA-2 region / site model outputs
    """
    region_id = scfg.Value(None, type=str, position=1, required=True, help='Region ID')
    previous_consolidated_output = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
            '''
            S3 path to consolidated regions / sites from previous
            iteration
            '''))
    current_output = scfg.Value(None, type=str, position=3, required=True, help=ub.paragraph(
            '''
            S3 path to regions / sites from current iteration
            '''))
    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands
            '''))
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            S3 Output directory for STAC item / asset egress
            '''))
    performer_suffix = scfg.Value(None, type=str, short_alias=['s'], help='Performer suffix if present, e.g. KIT')
    iou_threshold = scfg.Value(0.5, type=float, short_alias=['i'], help=ub.paragraph(
            '''
            IOU Threshold for determining duplicates(default: 0.5)
            '''))
    just_deconflict = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Don't copy previous sites, just deconflict current site IDs
            with respect to previous
            '''))


def main():
    config = PseudoliveConsolidateConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_pseudolive_consolidate(config)


def run_pseudolive_consolidate(config):

    from watch.cli.pseudolive_consolidate import pseudolive_consolidate
    if config.aws_profile is not None:
        aws_base_command =\
            ['aws', 's3', '--profile', config.aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    local_previous_dir = os.path.join('/tmp', 'previous_data')
    local_current_dir = os.path.join('/tmp', 'current_data')

    subprocess.run([*aws_base_command, '--recursive',
                    config.previous_consolidated_output, local_previous_dir],
                   check=True)

    # Quickly check and short-circuit if no "previous" data was copied
    # over (e.g. if there was no previous data)
    if len(os.listdir(local_previous_dir)) == 0:
        subprocess.run([*aws_base_command, '--recursive',
                        config.current_output, config.outbucket],
                       check=True)
        return

    subprocess.run([*aws_base_command, '--recursive',
                    config.current_output, local_current_dir],
                   check=True)

    local_consolidated_dir = os.path.join('/tmp', 'consolidated_out')
    pseudolive_consolidate(config.region_id,
                           local_previous_dir,
                           local_current_dir,
                           local_consolidated_dir,
                           config.iou_threshold,
                           performer_suffix=config.performer_suffix)

    subprocess.run([*aws_base_command, '--recursive',
                    local_consolidated_dir, config.outbucket],
                   check=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub
from torch.utils import data


if 1:
    # FIXME: put the vendored package into our namespace.
    try:
        PARENT_DPATH = ub.Path(__file__).parent
    except NameError:
        from watch.tasks.sam import predict as this_mod
        PARENT_DPATH = ub.Path(this_mod.__file__)
    TPL_DPATH = PARENT_DPATH / 'tpl'

    import sys
    import os
    sys.path.append(os.fspath(TPL_DPATH / 'segment_anything'))
    import segment_anything


class SAMConfig(scfg.DataConfig):
    input_kwcoco = scfg.Value(None, help='input kwcoco dataset')
    output_kwcoco = scfg.Value(None, help='output')

    weights_fpath = scfg.Value(None, help='path to pytorch model weights')

    window_dims = (1024, 1024)
    fixed_resolution = "10GSD"
    batch_size = 1
    window_overlap = 0.33333
    device = scfg.Value(0)
    select_images = None
    track_emissions = True
    data_workers = 2
    io_workers = 0
    assets_dname = 'sam'


def rgb_from_kwcoco_frame(frame):
    import torchvision.transforms.functional as F
    import kwimage
    import torch
    # import numpy as np
    modes = frame['modes']

    if 'red|green|blue' in modes:
        chw = modes['red|green|blue']
        is_nan = torch.isnan(chw)
        rgb_nan_frac = is_nan.sum() / is_nan.numel()
    else:
        rgb_nan_frac = 1.0

    # print('----')
    # print(f'rgb_nan_frac={rgb_nan_frac}')
    if rgb_nan_frac >= 0.0:
        # fallback on pan
        if 'pan' in modes:
            pan_chw = modes['pan']
            pan_is_nan = torch.isnan(pan_chw)
            pan_nan_frac = pan_is_nan.sum() / pan_is_nan.numel()
            # print(f'pan_nan_frac={pan_nan_frac}')
            if rgb_nan_frac > pan_nan_frac:
                pan_hwc = pan_chw.permute(1, 2, 0).cpu().numpy()
                pan_hwc = kwimage.atleast_3channels(pan_hwc)
                chw = torch.Tensor(pan_hwc).permute(2, 0, 1)

    hwc = chw.permute(1, 2, 0).cpu().numpy()
    normed_rgb = kwimage.normalize_intensity(hwc)
    chw = F.to_tensor(normed_rgb)
    return chw


class SAMWrapperDataset(data.Dataset):

    def __init__(self, subdset):
        self.subdset = subdset

    def __len__(self):
        return len(self.subdset)

    def __getitem__(self, index):
        import kwimage
        import torch
        import numpy as np

        torch_dataset = self.subdset

        item = torch_dataset[index]

        frame0 = item['frames'][0]
        chw = rgb_from_kwcoco_frame(frame0)
        hwc = chw.permute(1, 2, 0).cpu().numpy()
        norm01_resized, resize_info = kwimage.imresize(
            hwc, dsize=(1024, 1024), interpolation='lanczos',
            letterbox=True, border_value=np.nan, return_info=True)
        norm01_resized = norm01_resized.clip(0, 1)
        hwc_isnodata_mask = np.isnan(norm01_resized)

        # Get a nodata mask for the output we expect, Only mask an output block
        # if more than half of its input pixels were nodata.
        import einops
        factor = 16
        isnodat_mask = hwc_isnodata_mask.all(axis=2).astype(np.uint8)
        isnodata_blocks = einops.rearrange(isnodat_mask, '(h2 s1) (w2 s2) -> h2 w2 (s1 s2)', s1=factor, s2=factor)
        output_isnodata_mask = isnodata_blocks.sum(axis=2) > (factor * factor) // 2

        norm255_resized = (np.nan_to_num(norm01_resized) * 255).round().astype(np.uint8)
        input_image_torch = torch.as_tensor(norm255_resized)

        chw255 = input_image_torch.permute(2, 0, 1).contiguous()

        # coco_dset = torch_dataset.sampler.dset
        gid = frame0['gid']
        # coco_img = coco_dset.coco_image(gid)

        # SAM will always downsample the output by a factor of 16 after it has
        # been resized to 1024x1024. Munge the transforms to account for this.
        fakeoutspace_from_outspace = kwimage.Affine.scale(16)
        outspace_from_fakeoutspace = fakeoutspace_from_outspace.inv()

        # The dataloader is assuming the output will be the same shape as the
        # input but that's not true. We will call this assumed output space the
        # fake outspace. The real outspace is actually downsampled by a factor
        # of 16.

        fake_output_dsize = kwimage.Box.from_dsize(frame0['output_image_dsize'])
        fake_outspace_slice = kwimage.Box.from_slice(frame0['output_space_slice'])
        fakeoutspace_from_vid = kwimage.Affine.coerce(scale=frame0['scale_outspace_from_vid'])

        output_dsize = fake_output_dsize.warp(outspace_from_fakeoutspace).quantize()
        outspace_slice_ = fake_outspace_slice.warp(outspace_from_fakeoutspace)
        asset_slice = outspace_slice_.quantize().clip(0, 0, None, None).to_slice()
        asset_dsize = (output_dsize.width, output_dsize.height)

        vid_from_fakeoutspace = fakeoutspace_from_vid.inv()
        vid_from_outspace = vid_from_fakeoutspace @ fakeoutspace_from_outspace
        outspace_from_vid = vid_from_outspace.inv()

        resize_offset = resize_info['offset']
        resize_scale = resize_info['scale']
        undo_resize = kwimage.Affine.coerce(scale=resize_scale,
                                            offset=resize_offset).inv()
        if not undo_resize.isclose_identity():
            kwimage.warp_affine
            raise NotImplementedError('todo')

        item['chw255'] = chw255
        item['chw_isnodata_mask'] = torch.as_tensor(hwc_isnodata_mask.transpose(2, 1, 0)).contiguous()
        item['output_isnodata_mask'] = np.ascontiguousarray(output_isnodata_mask)
        item['gid'] = gid
        item['asset_slice'] = asset_slice
        item['asset_dsize'] = asset_dsize
        item['scale_asset_from_vid'] = outspace_from_vid.decompose()['scale']
        return item


class DenseFeaturePredictor:
    """
    Base class for computing per-image dense features
    """
    short_code = NotImplemented
    chan_code = NotImplemented
    proc_name = NotImplemented
    WrapperDsetClass = NotImplemented

    def __init__(self, config):
        self.config = config

    def setup(self):
        from watch.tasks.fusion.datamodules import kwcoco_datamodule
        from watch.tasks.fusion.coco_stitcher import CocoStitchingManager
        from watch.utils import util_parallel
        from watch.utils import process_context

        config = self.config
        datamodule = kwcoco_datamodule.KWCocoVideoDataModule(
            test_dataset=config.input_kwcoco,
            batch_size=config.batch_size,
            fixed_resolution=config.fixed_resolution,
            num_workers=config.data_workers,
            window_dims=config.window_dims,
            select_images=config.select_images,
            window_overlap=config.window_overlap,
            force_bad_frames=True,
            resample_invalid_frames=0,
            time_steps=1,
            channels="red|green|blue,pan",
        )
        datamodule.setup('test')
        self.datamodule = datamodule

        torch_dataset = self.datamodule.torch_datasets['test']
        coco_dset = torch_dataset.sampler.dset
        coco_dset.clear_annotations()
        self.out_dset = coco_dset.copy()
        self.out_dset.fpath = ub.Path(config.output_kwcoco)
        # TODO: graceful bundle changes
        self.out_dset.reroot(absolute=True)

        self.proc_context = process_context.ProcessContext(
            name=self.proc_name,
            config=dict(config),
            track_emissions=config.track_emissions,
        )

        # Not sure why targets are not in image-first order, but let's ensure
        # they are so we can fix it.
        torch_dataset.new_sample_grid['targets'] = sorted(
            torch_dataset.new_sample_grid['targets'],
            key=lambda x: (x['video_id'], x['main_gid']))

        self.wrapper_dset = self.WrapperDsetClass(torch_dataset)

        writer_queue = util_parallel.BlockingJobQueue(max_workers=config.io_workers)
        self.coco_stitcher = CocoStitchingManager(
            self.out_dset,
            short_code=self.short_code,
            chan_code=self.chan_code,
            # stiching_space='image',
            stiching_space='video',  # stitch in video space
            writer_queue=writer_queue,
            expected_minmax=(0, 1),
            assets_dname=config.assets_dname or self.short_code,
        )

        self.device = self.config.device
        sam_model = segment_anything.sam_model_registry['vit_h'](checkpoint=config.weights_fpath)
        sam_model = sam_model.eval()
        sam_model = sam_model.to(self.device)
        self.sam_model = sam_model

    def run(self):
        import rich
        from torch.utils import data as torch_data
        from watch.utils import util_progress
        from kwcoco.util import util_json
        import torch

        out_fpath = ub.Path(self.out_dset.fpath)
        pred_dpath = out_fpath.parent
        pred_dpath.ensuredir()

        self.proc_context.add_disk_info(pred_dpath)
        self.proc_context.start()

        loader = torch_data.DataLoader(
            self.wrapper_dset, batch_size=self.config.batch_size,
            num_workers=self.config.data_workers,
            shuffle=False, pin_memory=False,
            # worker_init_fn=worker_init_fn,
            collate_fn=ub.identity,  # disable collation
        )
        batch_iter = iter(loader)
        # torch.set_grad_enabled(False)

        pman = util_progress.ProgressManager()
        with pman, torch.no_grad():
            prog = pman.progiter(batch_iter, total=len(loader),
                                  desc=f'Predict {self.short_code}')
            for batch in prog:
                self.predict_batch(batch)

        for gid in self.coco_stitcher.managed_image_ids():
            self.coco_stitcher.submit_finalize_image(gid)
        self.coco_stitcher.writer_queue.wait_until_finished()

        obj = self.proc_context.stop()
        obj = util_json.ensure_json_serializable(obj)
        self.out_dset.dataset['info'].append(obj)
        self.out_dset._ensure_json_serializable()

        print(f'out_fpath={out_fpath}')
        rich.print(f'Pred Dpath: [link={pred_dpath}]{pred_dpath}[/link]')
        self.out_dset.dump(indent='    ')
        print('Finished predict')

    def predict_batch(self, batch):
        raise NotImplementedError


class SAMFeaturePredictor(DenseFeaturePredictor):
    short_code = 'sam'
    proc_name = 'watch.tasks.sam.predict'
    chan_code = 'sam.0:256'

    WrapperDsetClass = SAMWrapperDataset

    def __init__(self, config):
        super().__init__(config)

    def predict_batch(self, batch):
        import torch
        batch_chw255 = []
        batch_chw_isnodata = []
        for item in batch:
            chw255 = item['chw255']
            chw_isnodata = item['chw_isnodata_mask']
            batch_chw255.append(chw255)
            batch_chw_isnodata.append(chw_isnodata)
        bchw255 = torch.stack(batch_chw255, dim=0)
        bchw_isnodata = torch.stack(batch_chw_isnodata, dim=0)

        bchw_isnodata = bchw_isnodata.to(self.device)
        bchw255 = bchw255.to(self.device)
        input_image = self.sam_model.preprocess(bchw255)
        input_image[bchw_isnodata] = 0
        embedding_bchw = self.sam_model.image_encoder(input_image)

        feat_bchw = embedding_bchw.detach().cpu().numpy()

        for item, feat_chw in zip(batch, feat_bchw):
            data = feat_chw.transpose(1, 2, 0)
            gid = item['gid']
            asset_dsize = item['asset_dsize']
            asset_slice = item['asset_slice']

            # If the input was mostly nodata, mask the output
            output_isnodata_mask = item['output_isnodata_mask']
            data[output_isnodata_mask] = float('nan')

            scale_asset_from_vid = item['scale_asset_from_vid']
            self.coco_stitcher.accumulate_image(
                gid, asset_slice, data, asset_dsize=asset_dsize,
                scale_asset_from_stitchspace=scale_asset_from_vid)


def main(cmdline=1, **kwargs):
    """

    CommandLine:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)

    CommandLine:
        TEST_SAM=1 xdoctest -m watch.tasks.sam.predict main

    Example:
        >>> # xdoctest: +REQUIRES(env:TEST_SAM)
        >>> from watch.tasks.sam.predict import *  # NOQA
        >>> import watch
        >>> dvc_expt_dpath = watch.find_dvc_dpath(tags='phase2_expt', hardware='auto')
        >>> dvc_data_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> weights_fpath = dvc_expt_dpath / 'models/sam/sam_vit_h_4b8939.pth'
        >>> input_kwcoco = dvc_data_dpath / 'Drop6-MeanYear10GSD-V2/imgonly-KR_R001.kwcoco.zip'
        >>> kwargs = dict(weights_fpath=weights_fpath, input_kwcoco=input_kwcoco, output_kwcoco=(dvc_data_dpath / 'Drop6-MeanYear10GSD-V2/_test.kwcoco.zip'))
        >>> cmdline = 0
        >>> main(cmdline, **kwargs)

    Ignore:
        from watch.tasks.sam.predict import *  # NOQA
        import watch
        dvc_expt_dpath = watch.find_dvc_dpath(tags='phase2_expt', hardware='auto')
        dvc_data_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        weights_fpath = dvc_expt_dpath / 'models/sam/sam_vit_h_4b8939.pth'
        input_kwcoco = dvc_data_dpath / 'Drop6-MeanYear10GSD-V2/imganns-AE_R001.kwcoco.zip'
        kwargs = dict(weights_fpath=weights_fpath, input_kwcoco=input_kwcoco, output_kwcoco=(dvc_data_dpath / 'Drop6-MeanYear10GSD-V2/imganns-AE_R001_sam.kwcoco.zip'))
        cmdline = 0
        config = SAMConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
        import rich
        rich.print('config = ' + ub.urepr(config, nl=1))
        self = predictor = SAMFeaturePredictor(config)
        predictor.setup()
        # main(cmdline, **kwargs)

    Example:
        >>> # xdoctest: +SKIP
        >>> cmdline = 0
        >>> kwargs = dict()
        >>> main(cmdline=cmdline, **kwargs)
    """
    config = SAMConfig.cli(cmdline=cmdline, data=kwargs, strict=True)

    import rich
    rich.print('config = ' + ub.urepr(config, nl=1))

    predictor = SAMFeaturePredictor(config)
    predictor.setup()
    predictor.run()

    # if 0:
    #     import kwcoco
    #     coco_dset = kwcoco.CocoDataset.coerce(config.input_kwcoco)

    #     coco_img = coco_dset.images().coco_images[0]
    #     delayed = coco_img.imdelay(channels='red|green|blue')
    #     data = delayed.finalize()

    #     import kwimage
    #     import numpy as np
    #     norm01 = kwimage.normalize_intensity(data)
    #     norm255 = kwimage.ensure_uint255(np.nan_to_num(norm01))

    #     sam_model = segment_anything.sam_model_registry['vit_h'](checkpoint=config.weights_fpath)
    #     sam_model = sam_model.eval()
    #     device = 0
    #     sam_model = sam_model.to(device)

    #     import torch
    #     torch.set_grad_enabled(False)

    #     if 0:
    #         sam_predictor = segment_anything.SamPredictor(sam_model)
    #         sam_predictor.set_image(norm255)
    #         embedding_bchw_v1 = sam_predictor.get_image_embedding()
    #         embedding_hwc_v1 = embedding_bchw_v1[0].permute(1, 2, 0).detach().cpu().numpy()
    #         embedding_viz_v1 = kwimage.normalize_intensity(embedding_hwc_v1[..., 0:3])

    #     if 1:
    #         import torch
    #         norm01_resized, info = kwimage.imresize(
    #             norm01, dsize=(1024, 1024), interpolation='lanczos',
    #             letterbox=True, border_value=np.nan, return_info=True)
    #         norm01_resized = norm01_resized.clip(0, 1)
    #         # isnan_resized = np.isnan(norm01_resized)
    #         norm255_resized = (norm01_resized * 255)
    #         input_image_torch = torch.as_tensor(norm255_resized, device=sam_model.device)
    #         input_image_torch = input_image_torch.permute(2, 0, 1).contiguous()[None, :, :, :]
    #         input_image = sam_model.preprocess(input_image_torch)
    #         # input_image[:, :, 0:h, 0:w][nan_mask_bchw] = 0
    #         input_image.nan_to_num_()
    #         print(f'input_image_torch.shape={input_image_torch.shape}')
    #         print(f'input_image.shape={input_image.shape}')
    #         embedding_bchw = sam_model.image_encoder(input_image)
    #         embedding_hwc = embedding_bchw[0].permute(1, 2, 0).detach().cpu().numpy()
    #         embedding_viz = kwimage.normalize_intensity(embedding_hwc[..., 0:3])

    #     if 0:
    #         import kwplot
    #         kwplot.autompl()
    #         kwplot.figure(fnum=1, doclf=True)
    #         kwplot.imshow(norm255, pnum=(1, 3, 1))
    #         kwplot.imshow(embedding_viz, pnum=(1, 3, 2))
    #         kwplot.imshow(embedding_viz_v1, pnum=(1, 3, 3))


if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/watch/watch/tasks/sam/predict.py
        python -m predict

    Ignore:

        python -m watch.tasks.sam.predict \
            --input_kwcoco "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Drop6-MeanYear10GSD-V2/imganns-AE_R001.kwcoco.zip" \
            --output_kwcoco "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Drop6-MeanYear10GSD-V2/imganns-AE_R001_sam.kwcoco.zip" \
            --weights_fpath "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/sam/sam_vit_h_4b8939.pth" \
            --window_overlap=0.33333 \
            --data_workers="2" \
            --io_workers 0 \
            --assets_dname="teamfeats"

        smartwatch visualize /home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Drop6-MeanYear10GSD-V2/imganns-AE_R001_sam.kwcoco.zip \
            --channels "red|green|blue,pan,sam.0:3,sam.3:6,sam.6:9" --smart

    """
    main()



import os
import torch
import torch.nn as nn
from argparse import ArgumentParser, Namespace
from datetime import date
import pytorch_lightning as pl
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.callbacks import ModelCheckpoint, LearningRateMonitor

from .utils import setup_python_logging
from .models import UNet, UNet_blur
from .spacenet.datasets import S7_sort


class space7_sort(pl.LightningModule):
    def __init__(self, hparams):
        super().__init__()

        if isinstance(hparams, dict):
            hparams = Namespace(**hparams)

        self.criterion = nn.BCEWithLogitsLoss()

        self.crop_size = hparams.crop_size

        self.accuracy = pl.metrics.classification.Accuracy()

        self.save_hyperparameters(hparams)

        if self.hparams.backbone == 'unet':
            self.backbone = UNet(3, hparams.feature_dim)
        elif self.hparams.backbone == 'unet_blur':
            self.backbone = UNet_blur(3, hparams.feature_dim)

        self.classifier = self.head(2 * hparams.feature_dim)

        self.accuracy = pl.metrics.Accuracy()

    def head(self, in_channels):
        return nn.Sequential(
            #nn.Conv2d(in_channels, in_channels // 2, 7, bias=False, padding=3),
            # nn.ReLU(),
            #nn.BatchNorm2d(in_channels // 2),
            nn.Conv2d(in_channels, 1, 1, bias=False, padding=0),
        )

    def forward(self, image1, image2, date1, date2):
        image1 = self.backbone(image1)
        image2 = self.backbone(image2)

        return image1, image2, date1, date2

    def shared_step(self, batch):
        image1, image2, date1, date2 = batch['image1'], batch['image2'], batch['date1'], batch['date2']
        image1, image2, date1, date2 = self(image1, image2, date1, date2)
        prediction = self.classifier(torch.cat((image1, image2), dim=1))

        date1 = torch.stack(date1).permute(1, 0)
        date2 = torch.stack(date2).permute(1, 0)

        labels = torch.tensor([
            tuple(date1[x]) < tuple(date2[x]) for x in range(date1.shape[0])
        ]).float().cuda()
        labels = labels.unsqueeze(1).unsqueeze(1).repeat(1, self.crop_size, self.crop_size).unsqueeze(1)

        loss = self.criterion(prediction, labels)
        accuracy = self.accuracy((prediction > 0.), labels.int())

        output = {
            # 'prediction': prediction,
            # 'labels': labels,
            'accuracy': accuracy,
            'loss': loss,
        }
        return output

    def training_step(self, batch, batch_idx):
        output = self.shared_step(batch)
        self.log('acc', output['accuracy'])
        self.log('loss', output['loss'])
        return output

    def validation_step(self, batch, batch_idx):
        output = self.shared_step(batch)

        output = {key + "_val": val for key, val in output.items()}
        self.log('val_acc', output['accuracy_val'])
        self.log('val_loss', output['loss_val'])
        return output

    def train_dataloader(self):
        return torch.utils.data.DataLoader(
            S7_sort(train=True,
                    normalize=self.hparams.normalize,
                    crop_size=[self.hparams.crop_size, self.hparams.crop_size]),
            batch_size=self.hparams.batch_size,
            num_workers=self.hparams.workers
        )

    def val_dataloader(self):
        return torch.utils.data.DataLoader(
            S7_sort(train=False, normalize=self.hparams.normalize,
                    crop_size=[self.hparams.crop_size, self.hparams.crop_size]),
            batch_size=self.hparams.batch_size,
            num_workers=self.hparams.workers
        )

    def configure_optimizers(self):
        opt = torch.optim.AdamW(
            self.parameters(),
            lr=self.hparams.learning_rate)

        lr_scheduler = torch.optim.lr_scheduler.StepLR(
            opt, step_size=self.hparams.step_size,
            gamma=self.hparams.gamma)

        return {'optimizer': opt, 'lr_scheduler': lr_scheduler}


def main(args):
    if isinstance(args, dict):
        args = Namespace(**args)

    log_dir = '{}/{}/normalize_{}/{}'.format(
        args.save_dir,
        'space7_sort/',
        args.normalize,
        str(date.today()),
        )
    exp_name = 'default'
    logger = TensorBoardLogger(log_dir, name=exp_name)

    setup_python_logging(logger.log_dir)

    model = space7_sort(hparams=args)

    checkpoint_callback = ModelCheckpoint(monitor='val_acc', mode='max', save_top_k=2)
    lr_logger = LearningRateMonitor(logging_interval='step')

    trainer = pl.Trainer.from_argparse_args(args, logger=logger, callbacks=[checkpoint_callback, lr_logger])
    trainer.fit(model)


if __name__ == '__main__':
    parser = ArgumentParser()

    parser.add_argument('--max_epochs', type=int, default=100)
    parser.add_argument('--workers', type=int, default=8)
    parser.add_argument('--batch_size', type=int, default=32)
    parser.add_argument('--step_size', type=int, default=20)
    parser.add_argument('--learning_rate', type=float, default=.0003)
    parser.add_argument('--gamma', type=float, default=.1)
    parser.add_argument('--weight_decay', type=float, default=1e-5)
    parser.add_argument('--save_dir', type=str, default='logs/')
    parser.add_argument('--gpus', type=int)
    parser.add_argument('--feature_dim', type=int, default=64)
    parser.add_argument('--crop_size', type=int, default=128)
    parser.add_argument('--single_unet', action='store_true')
    parser.add_argument('--normalize', action='store_true')
    parser.add_argument('--backbone', help='choose from unet, unet_blur', default='unet_blur')
    parser.set_defaults(
        gpus=1,
        terminate_on_nan=True,
        check_val_every_n_epochs=1,
        log_every_n_steps=20,
        flush_logs_every_n_steps=20,
        single_unet=True,
        normalize=False
        )

    args = parser.parse_args()
    args.default_save_path = os.path.join(args.save_dir, "logs")

    main(args)



r"""
Prediction script for landcover features.

Given a checkout of the model and drop6 data, the following demos computing and
visualizing a subset of the features.

CommandLine:

    DVC_EXPT_DPATH=$(geowatch_dvc --tags=phase2_expt --hardware=auto)
    DVC_DATA_DPATH=$(geowatch_dvc --tags=phase2_data --hardware=auto)

    KWCOCO_BUNDLE_DPATH=$DVC_DATA_DPATH/Drop6
    DZYNE_LANDCOVER_MODEL_FPATH="$DVC_EXPT_DPATH/models/landcover/sentinel2.pt"

    INPUT_DATASET_FPATH=$KWCOCO_BUNDLE_DPATH/imganns-KR_R001.kwcoco.zip
    OUTPUT_DATASET_FPATH=$KWCOCO_BUNDLE_DPATH/imganns-KR_R001_landcover_small.kwcoco.zip

    echo "
    DVC_DATA_DPATH="$DVC_DATA_DPATH"
    DVC_EXPT_DPATH="$DVC_EXPT_DPATH"

    DZYNE_LANDCOVER_MODEL_FPATH="$DZYNE_LANDCOVER_MODEL_FPATH"

    INPUT_DATASET_FPATH="$INPUT_DATASET_FPATH"
    OUTPUT_DATASET_FPATH="$OUTPUT_DATASET_FPATH"
    "

    export CUDA_VISIBLE_DEVICES="1"
    python -m watch.tasks.landcover.predict \
        --dataset="$INPUT_DATASET_FPATH" \
        --deployed="$DZYNE_LANDCOVER_MODEL_FPATH"  \
        --device=0 \
        --num_workers=4 \
        --select_images='(.frame_index < 100) and (.sensor_coarse == "S2")' \
        --with_hidden=6 \
        --output="$OUTPUT_DATASET_FPATH"

    smartwatch stats $OUTPUT_DATASET_FPATH

    smartwatch visualize $OUTPUT_DATASET_FPATH \
        --animate=True --channels="red|green|blue,barren|forest|water,landcover_hidden.0:3,landcover_hidden.3:6" \
        --skip_missing=True --workers=4 --draw_anns=False --smart=True
"""
import torch
import datetime
import ubelt as ub
from pathlib import Path

import kwcoco
from torch.utils.data import DataLoader

from watch.utils import util_parallel
from watch.utils import util_progress
from . import detector
from .model_info import lookup_model_info
from .utils import setup_logging

import scriptconfig as scfg


class LandcoverPredictConfig(scfg.DataConfig):
    dataset = scfg.Value(None, required=True, help='input kwcoco dataset')
    deployed = scfg.Value(None, required=True, help='pytorch weights file')
    output = scfg.Value(None, required=True, help='output kwcoco dataset')
    num_workers = scfg.Value(0, type=str, help='number of dataloading workers. Can be "auto"')
    io_workers = scfg.Value('auto', type=str, help='Number of writer threads. Defaults to min(num_workers, 2)')
    device = scfg.Value('auto', type=str, help='auto, cpu, or integer of the device to use')
    select_images = scfg.Value(None, type=str, help='if specified, a jq operation to filter images')
    select_videos = scfg.Value(None, type=str, help='if specified, a jq operation to filter videos')
    with_hidden = scfg.Value(None, type=int, help='if true, also write out this many of the hidden activations')
    track_emissions = scfg.Value(True, help='Set to False to disable codecarbon')
    window_dim = scfg.Value(1024, help='Set to False to disable codecarbon')
    assets_dname = scfg.Value('_assets', help=ub.paragraph(
        '''
        The name of the top-level directory to write new assets.
        '''))


def predict(cmdline=1, **kwargs):
    """
    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> from watch.tasks.landcover.predict import *  # NOQA
        >>> from watch.tasks.landcover.predict import _predict_single
        >>> import kwcoco
        >>> import watch
        >>> dvc_data_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> dvc_expt_dpath = watch.find_dvc_dpath(tags='phase2_expt', hardware='auto')
        >>> dset = kwcoco.CocoDataset(dvc_data_dpath / 'Drop6/imganns-KR_R001.kwcoco.zip')
        >>> deployed = dvc_expt_dpath / 'models/landcover/sentinel2.pt'
        >>> kwargs = {
        >>>     'dataset': dset.fpath,
        >>>     'deployed': deployed,
        >>>     'output': ub.Path(dset.fpath).augment(stemsuffix='_landcover', multidot=True),
        >>>     'select_images': '.sensor_coarse == "S2"',
        >>> }
        >>> cmdline = 0
        >>> predict(cmdline, **kwargs)
    """
    from watch.utils.lightning_ext import util_device
    from watch.utils import process_context
    from watch.utils import kwcoco_extensions
    from watch.tasks.fusion.coco_stitcher import CocoStitchingManager
    import rich

    config = LandcoverPredictConfig.cli(cmdline=cmdline, data=kwargs, strict=True)

    rich.print('config = {}'.format(ub.urepr(config, align=':', nl=1)))

    coco_dset_filename = config.dataset
    weights_filename = Path(config.deployed)
    output_dset_filename = get_output_file(config.output)

    deployed = ub.Path(config.deployed)
    if not deployed.is_file():
        raise ValueError('Landcover model does not exist')

    device = util_device.coerce_devices(config.device)[0]
    print(f'device={device}')

    config.num_workers = util_parallel.coerce_num_workers(config.num_workers)
    if config.io_workers == 'auto':
        config.io_workers = min(2, config.num_workers)
    config.io_workers = util_parallel.coerce_num_workers(config.io_workers)

    input_dset = kwcoco.CocoDataset.coerce(coco_dset_filename)
    filtered_gids = kwcoco_extensions.filter_image_ids(
        input_dset, include_sensors=None, exclude_sensors=None,
        select_images=config.select_images, select_videos=config.select_videos)
    input_dset = input_dset.subset(filtered_gids)
    print('Selected input_dset = {!r}'.format(input_dset))

    model_info = lookup_model_info(weights_filename)
    ptdataset = model_info.create_dataset(input_dset)
    model = model_info.load_model(weights_filename, device)

    print('Using {}'.format(type(model_info).__name__))

    print('Creating output dataset')
    output_dset = input_dset.copy()
    output_dset.fpath = output_dset_filename

    print('Initialize dataloader')
    dataloader = DataLoader(ptdataset, num_workers=config.num_workers,
                            batch_size=None, collate_fn=lambda x: x)

    # Start the worker processes before we do threading
    print('Initialize dataloader iter')
    dataloader_iter = iter(dataloader)

    # Create a queue that writes data to disk in the background
    print('Initialize stitchers')
    writer_queue = util_parallel.BlockingJobQueue(max_workers=config.io_workers)

    landcover_stitcher = CocoStitchingManager(
        output_dset,
        'landcover',
        chan_code='|'.join(model_info.model_outputs),
        stiching_space='image',
        writer_queue=writer_queue,
        expected_minmax=(0, 1),
        assets_dname=config.assets_dname,
    )

    num_hidden = config.with_hidden
    if config.with_hidden:
        _register_hidden_layer_hook(model)

        hidden_stitcher = CocoStitchingManager(
            output_dset,
            'landcover_hidden',
            chan_code=f'landcover_hidden.0:{num_hidden}',
            stiching_space='image',
            writer_queue=writer_queue,
            assets_dname=config.assets_dname,
        )
        hidden_stitcher.num_hidden = num_hidden
    else:
        model._activation_cache = None
        hidden_stitcher = None

    print('Initialize process context')
    proc_context = process_context.ProcessContext(
        type='process',
        name='watch.tasks.invariants.predict',
        config=config.to_dict(),
        track_emissions=config.track_emissions,
    )
    proc_context.start()

    print('Starting main predict loop')
    # pman = util_progress.ProgressManager('progiter')
    pman = util_progress.ProgressManager('rich')

    window_dim = config.window_dim

    with pman, torch.no_grad():
        _prog = pman.progiter(dataloader_iter, total=len(dataloader),
                              desc='predict landcover')
        for img_info in _prog:
            try:
                _predict_single(
                    img_info, model, model_info.model_outputs,
                    landcover_stitcher, hidden_stitcher,
                    output_dset=output_dset, window_dim=window_dim)

            except KeyboardInterrupt:
                print('interrupted')
                break
            except Exception as ex:
                print('warning: ex = {}'.format(ub.urepr(ex, nl=1)))
                print('Unable to load id:{} - {}'.format(img_info['id'], img_info['name']))
                raise

    writer_queue.wait_until_finished()

    print('Finish process context')
    proc_context.add_disk_info(ub.Path(input_dset.fpath).parent)
    proc_context.add_device_info(device)
    proc_context.stop()
    output_dset.dataset['info'].append(proc_context.obj)

    output_dset.dump(str(output_dset_filename), indent=2)
    print('output written to {}'.format(output_dset_filename))


def _register_hidden_layer_hook(model):
    # TODO: generalize to other models
    # Specific to UNetR model
    # These are at half of the output image resolution.

    model._activation_cache = {}

    def record_hidden_activation(layer, input, output):
        activation = output.detach()
        model._activation_cache['hidden'] = activation

    layer_of_interest = model.decoder1[3]
    layer_of_interest._forward_hooks.clear()
    layer_of_interest.register_forward_hook(record_hidden_activation)


def _predict_single(img_info,
                    model,
                    model_outputs,
                    landcover_stitcher,
                    hidden_stitcher,
                    output_dset: kwcoco.CocoDataset, window_dim=1024):
    """
    Modifies the coco dataset inplace, returns the data that needs to be
    written to disk.
    """
    import kwarray
    import kwimage
    gid = img_info['id']
    img = img_info['imgdata']

    ## Hardcoded params
    window_dims = (window_dim, window_dim)
    window_overlap = 0.3
    hidden_scale = 0.5  # scale from raw predictions to hidden predictions

    image_box = kwimage.Box.from_dsize(img.shape[0:2][::-1])

    if hidden_stitcher is not None:
        hidden_box = image_box.scale(hidden_scale).astype(int)
        hidden_dsize = hidden_box.dsize

    # Need to run a sliding window so we can manage larger image
    slider = kwarray.SlidingWindow(img.shape[0:2], window_dims,
                                   overlap=window_overlap,
                                   keepbound=True,
                                   allow_overshoot=True)

    valid_pred = False
    for img_slice in slider:
        subimg = img[img_slice]
        pred = detector.run(model, subimg, img_info)

        if pred is None:
            continue
        else:
            valid_pred = True

        scale = img.shape[0] / img_info["height"]
        landcover_stitcher.accumulate_image(gid, img_slice, pred, asset_dsize=img.shape[0:2][::-1], scale_asset_from_stitchspace=scale)

        if hidden_stitcher is not None:
            # Lots of hardcoded things here.
            # Hack to unpad here.
            if 'hidden' not in model._activation_cache:
                print('WARNING: we expected hidden, but its not there')
            else:
                hidden_raw = model._activation_cache['hidden'].cpu().numpy()
                model._activation_cache.pop('hidden', None)
                hidden = hidden_raw[0].transpose(1, 2, 0)
                h, w = pred.shape[0:2]
                hidden = hidden[0:h // 2, 0:w // 2, 0:hidden_stitcher.num_hidden]

                output_box = kwimage.Box.from_slice(img_slice)
                hidden_box = output_box.scale(hidden_scale).astype(int)
                hidden_slice = hidden_box.to_slice()
                hidden_stitcher.accumulate_image(
                    gid, hidden_slice, hidden, asset_dsize=hidden_dsize,
                    scale_asset_from_stitchspace=hidden_scale)

    if valid_pred:
        landcover_stitcher.submit_finalize_image(gid)
        if hidden_stitcher is not None:
            hidden_stitcher.submit_finalize_image(gid)


def get_output_file(output):
    default_output_filename = 'out_{}.kwcoco.json'.format(datetime.datetime.now().strftime('%Y%m%d_%H%M%S'))
    if output is None:
        output_dir = Path('/tmp')
        return output_dir.joinpath(default_output_filename)
    else:
        output = Path(output)
        if output.is_dir():
            return output.joinpath(default_output_filename)
        else:
            return output


if __name__ == '__main__':
    setup_logging()
    predict()



import os  # NOQA
import sys  # NOQA

# Adds the "modules" subdirectory to the python path.
# See https://gitlab.kitware.com/smart/watch/-/merge_requests/148#note_1050127
# for discussion of how to refactor this in the future.
sys.path.append(os.path.join(os.path.dirname(__file__), 'modules'))



"""
This is step 1 / 4 in predict.py

SeeAlso:

    predict.py

    prepare_kwcoco.py *

    tile_processing_kwcoco.py

    export_cold_result_kwcoco.py

    assemble_cold_result_kwcoco.py

This is a proof-of-concept for converting kwcoco files into the
expected data structures for pycold.

Relevant functions:
    * grab_demo_kwcoco_dataset - downloads a small kwcoco dataset for testing
    * stack_kwcoco - runs the stacking process on an entire kwcoco file
    * process_one_coco_image - runs the stacking for a single coco image.

Limitations:
    * Currently only handles Landsat-8

    * The quality bands are not exactly what I was expecting them to be,
      some of the quality filtering is stubbed out or disabled.

    * Not setup for an HPC environment yet, but that extension shouldn't be too
      hard.

    * Nodata values are currently not masked or handled

    * Configurations are hard-coded

TODO:
    - [ ] Incorporate watch/tasks/fusion/datamodules/qa_bands.py
"""
import kwcoco
import json
import numpy as np
import einops
import functools
import operator
import ubelt as ub
import itertools as it
import logging
from datetime import datetime as datetime_cls
import numpy as geek
import scriptconfig as scfg

logger = logging.getLogger(__name__)


try:
    from xdev import profile
except ImportError:
    from ubelt import identity as profile


class PrepareKwcocoConfig(scfg.DataConfig):
    """
    The docstring will be the description in the CLI help
    """

    coco_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        a path to a file to input kwcoco file
        '''))
    out_dpath = scfg.Value(None, help='output directory for the output')
    sensors = scfg.Value('L8', type=str, help='sensor type, default is "L8"')
    adj_cloud = scfg.Value(False, help='How to treat QA band, default is False: ignoring adj. cloud class')
    method = scfg.Value(None, help=ub.paragraph(
        '''
        stacking mode for original COLD or Hybrid, default is None, if
        HybridCOLD then stacked data include g, r, nir, swir16, swir22, ASI,
        tir, QA
        '''))
    resolution = scfg.Value('30GSD', help='if specified then data is processed at this resolution')
    workers = scfg.Value(0, help='number of parallel workers')


# TODO:
# For each sensor, register the specific bands we are interested in.
# This demo currently assumes landsat8
SENSOR_TO_INFO = {}
SENSOR_TO_INFO['L8'] = {
    'sensor_name': 'Landsat-8',
    'intensity_channels': 'blue|green|red|nir|swir16|swir22|lwir11',
    'quality_channels': 'quality',
    'quality_interpretation': 'FMASK'
}  # The name of quality_channels for Drop 4 is 'cloudmask'.

SENSOR_TO_INFO['S2'] = {
    'sensor_name': 'Sentinel-2',
    'intensity_channels': 'blue|green|red|nir|swir16|swir22|lwir11',
    'quality_channels': 'quality',
    'quality_interpretation': 'FMASK'
}

# Register different quality bit standards.
QA_INTERPRETATIONS = {}

# These are specs for TA1 processed data
QA_BIT = {
    'clear': 1 << 0,
    'cloud': 1 << 1,
    'cloud_adj': 1 << 2,
    'shadow': 1 << 3,
    'snow': 1 << 4,
    'water': 1 << 5,
}

QA_INTERPRETATIONS['FMASK'] = {
    'clear': 0,
    'water': 1,
    'cloud_shadow': 2,
    'snow': 3,
    'cloud': 4,
    'no_obs': 255,
}
QUALITY_BIT_INTERPRETATIONS = {}


@profile
def prepare_kwcoco_main(cmdline=1, **kwargs):
    """_summary_

    Args:
        cmdline (int, optional): _description_. Defaults to 1.

    Ignore:
        python -m watch.tasks.cold.prepare_kwcoco --help
        TEST_COLD=1 xdoctest -m watch.tasks.cold.prepare_kwcoco prepare_kwcoco_main

    Example:
        >>> # xdoctest: +REQUIRES(env:TEST_COLD)
        >>> from watch.tasks.cold.prepare_kwcoco import prepare_kwcoco_main
        >>> from watch.tasks.cold.prepare_kwcoco import *
        >>> kwargs= dict(
        >>>   coco_fpath = ub.Path('/home/jws18003/data/dvc-repos/smart_data_dvc/Drop6/data_vali_split1_KR_R001.kwcoco.json'),
        >>>   out_dpath = ub.Path.appdir('/gpfs/scratchfs1/zhz18039/jws18003/kwcoco'),
        >>>   sensors = 'L8,S2',
        >>>   adj_cloud = False,
        >>>   method = None,
        >>> )
        >>> cmdline=0
        >>> prepare_kwcoco_main(cmdline, **kwargs)
    """
    pman = kwargs.pop('pman', None)
    config = PrepareKwcocoConfig.cli(cmdline=cmdline, data=kwargs)
    coco_fpath = config['coco_fpath']
    dpath = ub.Path(config['out_dpath']).ensuredir()
    sensors = config['sensors']
    adj_cloud = config['adj_cloud']
    method = config['method']
    out_dir = dpath / 'stacked'
    workers = config['workers']
    resolution = config.resolution
    meta_fpath = stack_kwcoco(coco_fpath, out_dir, sensors, adj_cloud, method,
                              pman, workers, resolution)
    return meta_fpath


# function for decoding HLS qa band
def qa_decoding(qa_array):
    """
    This function is modified from qabitval_array_HLS function
    (https://github.com/GERSL/pycold/blob/c5b380eccc2916e5c3aec0bbd2b1982e114b75b1/src/python/pycold/imagetool/prepare_ard.py#L74)
    """
    unpacked = np.full(qa_array.shape, QA_INTERPRETATIONS['FMASK']['clear'])

    QA_CLOUD_unpacked = geek.bitwise_and(qa_array, QA_BIT['cloud'])
    QA_CLOUD_ADJ = geek.bitwise_and(qa_array, QA_BIT['cloud_adj'])
    QA_SHADOW_unpacked = geek.bitwise_and(qa_array, QA_BIT['shadow'])
    QA_SNOW_unpacked = geek.bitwise_and(qa_array, QA_BIT['snow'])
    QA_WATER_unpacked = geek.bitwise_and(qa_array, QA_BIT['water'])

    unpacked[QA_WATER_unpacked > 0] = QA_INTERPRETATIONS['FMASK']['water']
    unpacked[QA_SNOW_unpacked > 0] = QA_INTERPRETATIONS['FMASK']['snow']
    unpacked[QA_SHADOW_unpacked >
             0] = QA_INTERPRETATIONS['FMASK']['cloud_shadow']
    unpacked[QA_CLOUD_ADJ > 0] = QA_INTERPRETATIONS['FMASK']['cloud']
    unpacked[QA_CLOUD_unpacked > 0] = QA_INTERPRETATIONS['FMASK']['cloud']
    unpacked[qa_array == QA_INTERPRETATIONS['FMASK']
             ['no_obs']] = QA_INTERPRETATIONS['FMASK']['no_obs']

    return unpacked


def qa_decoding_no_boundary(qa_array):
    """
    This function is modified from qabitval_array_HLS function
    (https://github.com/GERSL/pycold/blob/c5b380eccc2916e5c3aec0bbd2b1982e114b75b1/src/python/pycold/imagetool/prepare_ard.py#L74)
    """
    unpacked = np.full(qa_array.shape, QA_INTERPRETATIONS['FMASK']['clear'])

    QA_CLOUD_unpacked = geek.bitwise_and(qa_array, QA_BIT['cloud'])
    QA_CLOUD_ADJ = geek.bitwise_and(qa_array, QA_BIT['cloud_adj'])
    QA_SHADOW_unpacked = geek.bitwise_and(qa_array, QA_BIT['shadow'])
    QA_SNOW_unpacked = geek.bitwise_and(qa_array, QA_BIT['snow'])
    QA_WATER_unpacked = geek.bitwise_and(qa_array, QA_BIT['water'])

    unpacked[QA_WATER_unpacked > 0] = QA_INTERPRETATIONS['FMASK']['water']
    unpacked[QA_SNOW_unpacked > 0] = QA_INTERPRETATIONS['FMASK']['snow']
    unpacked[QA_SHADOW_unpacked >
             0] = QA_INTERPRETATIONS['FMASK']['cloud_shadow']
    unpacked[QA_CLOUD_unpacked > 0] = QA_INTERPRETATIONS['FMASK']['cloud']
    unpacked[QA_CLOUD_ADJ > 0] = QA_INTERPRETATIONS['FMASK']['clear']
    unpacked[qa_array == QA_INTERPRETATIONS['FMASK']
             ['no_obs']] = QA_INTERPRETATIONS['FMASK']['no_obs']

    return unpacked


def setup_logging():
    # TODO: handle HPC things here in addition to stdout for doctests
    logging.basicConfig(level='INFO')


##########################################################################
#                   Functions for Artificial Surface Index (ASI)                   #
#  See original code: https://github.com/GERSL/ASI_py/blob/main/ASI_standalone.py  #
##########################################################################

def hist_cut(band, mask, fill_value=-9999, k=3, minmax='std'):
    if minmax == 'std':
        mean = band[mask].mean()
        std = band[mask].std()
        low_val = (mean - k * std)
        high_val = (mean + k * std)
    else:
        low_val, high_val = minmax  # use specified value range.
    is_low = band < low_val
    is_high = band > high_val
    mask_invalid_index = is_low | is_high
    band[mask_invalid_index] = fill_value
    return band, ~mask_invalid_index


def minmax_norm(band, mask, fill_value=-9999):
    max_val = band[mask].max()
    min_val = band[mask].min()
    extent = max_val - min_val
    if extent != 0:
        shifted = band - min_val
        scaled = shifted / extent
        band[mask] = scaled[mask]
    band[~mask] = fill_value
    return band


# Artificial Surface Index (ASI) is designed based the surface reflectance
# imagery of Landsat 8.
def artificial_surface_index(
        Blue, Green, Red, NIR, SWIR1, SWIR2, Scale, MaskValid_Obs, fillV):
    # The calculation chain.

    # Artificial surface Factor (AF).
    AF = (NIR - Blue) / (NIR + Blue) + 0.000001
    AF, MaskValid_AF = hist_cut(AF, MaskValid_Obs, fillV, 6, [-1, 1])
    MaskValid_AF_U = MaskValid_AF & MaskValid_Obs
    AF_Norm = minmax_norm(AF, MaskValid_AF_U, fillV)

    # Vegetation Suppressing Factor (VSF).
    # Modified Soil Adjusted Vegetation Index (MSAVI).
    MSAVI = ((2 * NIR + 1 * Scale) -
             np.sqrt((2 * NIR + 1 * Scale)**2 - 8 * (NIR - Red))) / 2
    MSAVI, MaskValid_MSAVI = hist_cut(MSAVI, MaskValid_Obs, fillV, 6, [-1, 1])
    NDVI = (NIR - Red) / (NIR + Red) + 0.000001
    NDVI, MaskValid_NDVI = hist_cut(NDVI, MaskValid_Obs, fillV, 6, [-1, 1])
    VSF = 1 - MSAVI * NDVI
    MaskValid_VSF = MaskValid_MSAVI & MaskValid_NDVI & MaskValid_Obs
    VSF_Norm = minmax_norm(VSF, MaskValid_VSF, fillV)

    # Soil Suppressing Factor (SSF).
    # Derive the Modified Bare soil Index (MBI).
    MBI = (SWIR1 - SWIR2 - NIR) / (SWIR1 + SWIR2 + NIR) + 0.5
    MBI, MaskValid_MBI = hist_cut(MBI, MaskValid_Obs, fillV, 6, [-0.5, 1.5])
    # Deriving Enhanced-MBI based on MBI and MNDWI.
    MNDWI = (Green - SWIR1) / (Green + SWIR1) + 0.000001
    MNDWI, MaskValid_MNDWI = hist_cut(MNDWI, MaskValid_Obs, fillV, 6, [-1, 1])
    EMBI = ((MBI + 0.5) - (MNDWI + 1)) / ((MBI + 0.5) + (MNDWI + 1))
    EMBI, MaskValid_EMBI = hist_cut(EMBI, MaskValid_Obs, fillV, 6, [-1, 1])
    # Derive SSF.
    SSF = (1 - EMBI)
    MaskValid_SSF = MaskValid_MBI & MaskValid_MNDWI & MaskValid_EMBI & MaskValid_Obs
    SSF_Norm = minmax_norm(SSF, MaskValid_SSF, fillV)

    # Modulation Factor (MF).
    MF = (Blue + Green - NIR - SWIR1) / (Blue + Green + NIR + SWIR1) + 0.000001
    MF, MaskValid_MF = hist_cut(MF, MaskValid_Obs, fillV, 6, [-1, 1])
    MaskValid_MF_U = MaskValid_MF & MaskValid_Obs
    MF_Norm = minmax_norm(MF, MaskValid_MF_U, fillV)

    # Derive Artificial Surface Index (ASI).
    ASI = AF_Norm * SSF_Norm * VSF_Norm * MF_Norm
    MaskValid_ASI = MaskValid_AF_U & MaskValid_VSF & MaskValid_SSF & MaskValid_MF_U & MaskValid_Obs
    ASI[~MaskValid_ASI] = fillV

    return ASI


def stack_kwcoco(coco_fpath, out_dir, sensors, adj_cloud, method, pman=None,
                 workers=0, resolution=None):
    """
    Args:
        coco_fpath (str | PathLike | CocoDataset):
            the kwcoco dataset to convert

        out_dir (str | PathLike): path to write the data

    Returns:
        List[Dict]: a list of dictionary result objects

    Example:
        >>> # xdoctest: +SKIP
        >>> # TODO: readd this doctest
        >>> from pycold.imagetool.prepare_kwcoco import *  # NOQA
        >>> setup_logging()
        >>> coco_dset = watch.coerce_kwcoco('watch-msi')
        >>> coco_fpath = coco_dset.fpath
        >>> #coco_fpath = grab_demo_kwcoco_dataset()
        >>> dpath = ub.Path.appdir('pycold/tests/stack_kwcoco').ensuredir()
        >>> out_dir = dpath / 'stacked'
        >>> results = stack_kwcoco(coco_fpath, out_dir)
    """
    # TODO: configure
    out_dir = ub.Path(out_dir)

    # Load the kwcoco dataset
    dset = kwcoco.CocoDataset.coerce(coco_fpath)

    # Get all images ids sorted in temporal order per video
    all_images = dset.images(list(ub.flatten(dset.videos().images)))

    # For now, it supports only L8
    flags = [s in {'L8'} for s in all_images.lookup('sensor_coarse')]
    # flags = [s in sensors for s in all_images.lookup('sensor_coarse')]
    all_images = all_images.compress(flags)

    if len(all_images) == 0:
        raise Exception('No images!')

    image_id_iter = iter(all_images)

    # For now, it supports only L8
    jobs = ub.JobPool(mode='process', max_workers=workers)
    if pman is not None:
        image_id_iter = pman.progiter(image_id_iter, desc='submit prepare jobs', transient=True)
    for image_id in image_id_iter:
        coco_image: kwcoco.CocoImage = dset.coco_image(image_id)
        coco_image = coco_image.detach()
        # Transform the image data into the desired block structure.
        jobs.submit(process_one_coco_image,
                    coco_image, out_dir, adj_cloud, method, resolution)

    job_iter = jobs.as_completed()
    if pman is not None:
        job_iter = pman.progiter(job_iter, desc='collect prepare jobs')

    for job in job_iter:
        meta_fpath = job.result()

    return meta_fpath


@profile
def process_one_coco_image(coco_image, out_dir, adj_cloud, method, resolution):
    """
    Args:
        coco_image (kwcoco.CocoImage): the image to process

        out_dir (Path): path to write the image data

        resolution (str | None): resolution to process at (e.g. 30GSD).

    Returns:
        Dict: result dictionary with keys:
            status (str) : either a string passed or failed
            fpaths (List[str]): a list of files that were written
    """
    n_block_x = 20
    n_block_y = 20
    is_partition = True  # hard coded

    # Use the COCO name as a unique filename id.
    image_name = coco_image.img.get('name', None)
    video_name = coco_image.video.get('name', None)
    if image_name is None:
        image_name = 'img_{:06d}'.format(coco_image.img['id'])
    if video_name is None:
        video_name = 'vid_{:06d}'.format(coco_image.video['id'])

    # Preconstruct the file names used in the inner loops
    fname_json = (image_name + '.json')
    fname_npy = (image_name + '.npy')

    video_dpath = (out_dir / video_name).ensuredir()

    # Other relevant coco metadata
    date_captured = coco_image.img['date_captured']
    ordinal_date = datetime_cls.strptime(
        date_captured[:10], '%Y-%m-%d').toordinal()
    # frame_index = coco_image.img['frame_index']
    n_cols = coco_image.img['width']
    n_rows = coco_image.img['height']
    # Determine what sensor the image is from.
    # Note: if kwcoco needs to register more fine-grained sensor
    # information we can do that.
    sensor = coco_image.img['sensor_coarse']
    # assert sensor == 'L8', 'MWE only supports landsat-8 for now'

    # Given the sensor, determine what the intensity and quality band
    # we should request are.
    sensor_info = SENSOR_TO_INFO[sensor]
    intensity_channels = sensor_info['intensity_channels']
    quality_channels = sensor_info['quality_channels']
    quality_interpretation = sensor_info['quality_interpretation']
    quality_bits = QA_INTERPRETATIONS[quality_interpretation]
    # Specify how we are going to handle spatial resampling and nodata
    delay_kwargs = {
        'nodata_method': None,
        'space': 'video',
        'resolution': resolution,
    }

    # Construct delayed images. These represent a tree of image
    # operations that will resample the image at the desired resolution
    # as well as align it with other images in the sequence.
    # NOTE: Issue occurs when setting resolution argument in imdelay
    delayed_im = coco_image.imdelay(channels=intensity_channels, **delay_kwargs)
    delayed_qa = coco_image.imdelay(channels=quality_channels, **delay_kwargs)

    # Check what shape the data would be loaded with if we finalized right now.
    h, w = delayed_im.shape[0:2]
    # Determine if padding is necessary to properly break the data into blocks.
    padded_w = int(np.ceil(w / n_block_x) * n_block_x)
    padded_h = int(np.ceil(h / n_block_y) * n_block_y)

    if padded_w != w or padded_h != h:
        # cropping using an oversized slice with clip=False and wrap=False is
        # equivalent to padding. In the future a more efficient pad operation
        # where the padding value can be specified will be added, but this will
        # work well enough for now.
        slice_ = (slice(0, padded_h), slice(0, padded_w))
        delayed_im = delayed_im.crop(slice_, clip=False, wrap=False)
        delayed_qa = delayed_qa.crop(slice_, clip=False, wrap=False)

    # It is important that the categorical QA band is not interpolated or
    # antialiased, whereas the intensity bands should be.
    qa_data = delayed_qa.finalize(interpolation='nearest', antialias=False)
    # Decoding QA band
    if adj_cloud:
        qa_unpacked = qa_decoding(qa_data)
    else:
        qa_unpacked = qa_decoding_no_boundary(qa_data)

    # First check the quality bands before loading all of the image data.
    # FIXME: the quality bits in this example are wrong.
    # Setting the threshold to zero to bypass for now.
    clear_threshold = 0
    if clear_threshold > 0:
        clear_bits = functools.reduce(
            operator.or_, ub.take(quality_bits, ['clear_land', 'clear_water']))
        noobs_bits = functools.reduce(
            operator.or_, ub.take(quality_bits, ['no_observation']))
        is_clear = (qa_data & clear_bits) > 0
        is_noobs = (qa_data & noobs_bits) > 0
        is_obs = ~is_noobs
        is_obs_clear = is_clear & is_obs
        clear_ratio = is_obs_clear.sum() / is_obs.sum()
    else:
        clear_ratio = 1

    if clear_ratio <= clear_threshold:
        logger.warn('Not enough clear observations for {}/{}'.format(
            video_name, image_name))
        return

    im_data = delayed_im.finalize(interpolation='cubic', antialias=True)

    # Zero padding affects margin block not to process COLD, leading missing npy in the second step.
    # To avoid this issue, filled with fake data in the padding area.
    # This will fix predict script running forever without raising error.
    # This issue will be addressed in export_cold_result_kwcoco.py, not here
    # padding_value_col = im_data[:, :padded_w - w]
    # im_data[:, w:] = padding_value_col
    # padding_value_row = im_data[:padded_h - h, :]
    # im_data[h:] = padding_value_row

    if method == 'ASI':
        Scale = 10000
        fill_value = 0
        B1 = im_data[:, :, 0]
        B2 = im_data[:, :, 1]
        B3 = im_data[:, :, 2]
        B4 = im_data[:, :, 3]
        B5 = im_data[:, :, 4]
        B6 = im_data[:, :, 5]
        MaskValid_Obs = ((B1 > 0) & (B1 < 1 * Scale) &
                         (B2 > 0) & (B2 < 1 * Scale) &
                         (B3 > 0) & (B3 < 1 * Scale) &
                         (B4 > 0) & (B4 < 1 * Scale) &
                         (B5 > 0) & (B5 < 1 * Scale) &
                         (B6 > 0) & (B6 < 1 * Scale)
                         )

        # Calculating ASI
        ASI = artificial_surface_index(
            B1.astype(
                np.float32), B2.astype(
                np.float32), B3.astype(
                np.float32), B4.astype(
                    np.float32), B5.astype(
                        np.float32), B6.astype(
                            np.float32), Scale, MaskValid_Obs, fill_value)
        # Get land mask.
        MNDWI = (B2 - B5) / (B2 + B5)
        MNDWI, MaskValid_MNDWI = hist_cut(
            MNDWI, MaskValid_Obs, fill_value, 6, [-1, 1])
        # Water threshold for MNDWI (may need to be adjusted for different
        # study areas).
        Water_Th = 0
        MaskLand = (MNDWI < Water_Th)

        # Convert dtype from float32 to int16
        ASI = ASI * Scale
        ASI = ASI.astype('int16')
        ASI[ASI == 0] = fill_value

        # Exclude water pixels.
        ASI[~MaskLand] = fill_value
        ASI = ASI.reshape(ASI.shape[0], ASI.shape[1], 1)
        false_band = np.full((ASI.shape[0], ASI.shape[1], 1), 0)
        # input for Hybrid-COLD (with ASI) = B2, B3, B4, B5, B6, ASI
        data = np.concatenate(
            [im_data[:, :, 1:6], ASI, false_band, qa_unpacked], axis=2)

    else:
        data = np.concatenate([im_data, qa_unpacked], axis=2)

    result_fpaths = []

    metadata = {
        'image_name': image_name,
        'date_captured': date_captured,
        'ordinal_date': ordinal_date,
        'region_id': video_name,
        'n_cols': n_cols,
        'n_rows': n_rows,
        'video_w': w,
        'video_h': h,
        'padded_n_cols': padded_w,
        'padded_n_rows': padded_h,
        'n_block_x': n_block_x,
        'n_block_y': n_block_y,
        'adj_cloud': adj_cloud,
        'method': method
    }

    if is_partition:
        bw = int(padded_w / n_block_x)  # width of a block
        bh = int(padded_h / n_block_y)  # height of a block

        # Use einops to rearrange the data into blocks
        # Question: would using numpy strided tricks be faster?
        blocks = einops.rearrange(
            data, '(nby bh) (nbx bw) c -> nbx nby bh bw c', bw=bw, bh=bh)

        # FIXME: Disable skipping until QA bands are handled correctly
        SKIP_BLOCKS_WITH_QA = False

        for i, j in it.product(range(n_block_y), range(n_block_x)):
            block = blocks[i, j]
            bh, bw = block.shape[0:2]

            if SKIP_BLOCKS_WITH_QA:
                # check if no valid pixels in the chip, then eliminate
                qa_unique = np.unique(block[..., -1])
                qa_unique
                # skip blocks are all cloud, shadow or filled values in DHTC,
                # we also don't need to save pixel that has qa value of
                # 'QA_CLOUD', 'QA_SHADOW', or FILLED value (255)
                if ... and False:
                    continue

            block_dname = 'block_x{}_y{}'.format(i + 1, j + 1)
            block_dpath = video_dpath / block_dname
            block_fpath = block_dpath / fname_npy

            metadata.update({
                'x': i + 1,
                'y': j + 1,
                'total_pixels': bw * bh,
                'total_bands': int(block.shape[-1]),
            })
            if not block_fpath.exists():
                block_dpath.mkdir(exist_ok=True)
                meta_fpath = block_dpath / fname_json
                meta_fpath.write_text(json.dumps(metadata))
                np.save(block_fpath, block)
                result_fpaths.append(block_fpath)
                result_fpaths.append(meta_fpath)
        logger.info(
            'Stacked blocked image {}/{}'.format(video_name, image_name))
    else:
        # FIXME: seems broken. Comment on why and raise a NotImplementedError
        # if this is true.

        metadata.update({
            'total_pixels': int(np.prod(data.shape[0:2])),
            'total_bands': int(data.shape[-1]),
        })

        full_fpath = video_dpath / fname_npy
        if not block_fpath.exists():
            meta_fpath = video_dpath / fname_json
            meta_fpath.write_text(json.dumps(metadata))
            np.save(full_fpath, data)
            result_fpaths.append(full_fpath)
            result_fpaths.append(meta_fpath)
            logger.info(
                'Stacked full image {}/{}'.format(video_name, image_name))

    # FIXME: It seems strange that we return one of the metadata paths for just
    # one of the blocks instead of having a metadata file for all blocks.
    meta_fpath = block_dpath / (image_name + '.json')
    return meta_fpath


if __name__ == '__main__':
    prepare_kwcoco_main()



import datetime as datetime_mod
import numbers
import dateutil
import time
import ubelt as ub
import math
from datetime import datetime as datetime_cls

try:
    from xdev import profile
except ImportError:
    profile = ub.identity


class TimeValueError(ValueError):
    ...


class TimeTypeError(TypeError):
    ...


def isoformat(dt, sep='T', timespec='seconds', pathsafe=True):
    """
    A path-safe version of datetime_cls.isotime() that returns a
    path-friendlier version of a ISO 8601 timestamp.

    Args:
        dt (datetime_cls): datetime to format
        pathsafe (bool):

    References:
        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes

    SeeAlso:
        ubelt.timestamp

    Example:
        >>> from watch.utils.util_time import *  # NOQA
        >>> import datetime
        >>> items = []
        >>> dt = datetime_cls.now()
        >>> dt = ensure_timezone(dt, datetime_mod.timezone(datetime_mod.timedelta(hours=+5)))
        >>> items.append(dt)
        >>> dt = datetime_cls.utcnow()
        >>> items.append(dt)
        >>> dt = dt.replace(tzinfo=datetime_mod.timezone.utc)
        >>> items.append(dt)
        >>> dt = ensure_timezone(datetime_cls.now(), datetime_mod.timezone(datetime_mod.timedelta(hours=-5)))
        >>> items.append(dt)
        >>> dt = ensure_timezone(datetime_cls.now(), datetime_mod.timezone(datetime_mod.timedelta(hours=+5)))
        >>> items.append(dt)
        >>> print('items = {!r}'.format(items))
        >>> for dt in items:
        >>>     print('----')
        >>>     print('dt = {!r}'.format(dt))
        >>>     # ISO format is cool, but it doesnt give much control
        >>>     print(dt.isoformat())
        >>>     # Need a better version
        >>>     print(isoformat(dt))
        >>>     print(isoformat(dt, pathsafe=False))
    """
    if not pathsafe:
        return dt.isoformat(sep=sep, timespec=timespec)

    date_fmt = '%Y%m%d'
    if timespec == 'seconds':
        time_tmf = '%H%M%S'
    else:
        raise NotImplementedError(timespec)

    text = dt.strftime(''.join([date_fmt, sep, time_tmf]))
    if dt.tzinfo is not None:
        off = dt.utcoffset()
        off_seconds = off.total_seconds()
        if off_seconds == 0:
            # TODO: use codes for offsets to remove the plus sign if possible
            suffix = 'Z'
        elif off_seconds % 3600 == 0:
            tz_hour = int(off_seconds // 3600)
            suffix = '{:02d}'.format(tz_hour) if tz_hour < 0 else '+{:02d}'.format(tz_hour)
        else:
            suffix = _format_offset(off)
        text += suffix
    return text


def _format_offset(off):
    """
    Taken from CPython:
        https://github.com/python/cpython/blob/main/Lib/datetime_mod.py
    """
    s = ''
    if off is not None:
        if off.days < 0:
            sign = "-"
            off = -off
        else:
            sign = "+"
        hh, mm = divmod(off, datetime_mod.timedelta(hours=1))
        mm, ss = divmod(mm, datetime_mod.timedelta(minutes=1))
        s += "%s%02d:%02d" % (sign, hh, mm)
        if ss or ss.microseconds:
            s += ":%02d" % ss.seconds

            if ss.microseconds:
                s += '.%06d' % ss.microseconds
    return s


@profile
def coerce_datetime(data, default_timezone='utc', nan_policy='return-None',
                    none_policy='return-None'):
    """
    Parses a timestamp and always returns a timestamp with a timezone.
    If only a date is specified, the time is defaulted to 00:00:00
    If one is not discoverable a specified default is used.
    A nan or None input depends on nan_policy and none_policy.

    Args:
        data (None | str | datetime.datetime | datetime.date)

        default_timezone (str): defaults to utc.

        none_policy (str):
            How None inputs are handled. Can be:
                'cast-to-none': returns None
                'raise': raises an error

        nan_policy (str):
            How nan inputs are handled. Can be:
                'cast-to-none': returns None
                'raise': raises an error

    Returns:
        datetime.datetime | None

    Example:
        >>> from watch.utils.util_time import *  # NOQA
        >>> assert coerce_datetime(None) is None
        >>> assert coerce_datetime(float('nan')) is None
        >>> assert coerce_datetime('2020-01-01') == datetime_cls(2020, 1, 1, 0, 0, tzinfo=datetime_mod.timezone.utc)
        >>> assert coerce_datetime(datetime_cls(2020, 1, 1, 0, 0)) == datetime_cls(2020, 1, 1, 0, 0, tzinfo=datetime_mod.timezone.utc)
        >>> assert coerce_datetime(datetime_cls(2020, 1, 1, 0, 0).date()) == datetime_cls(2020, 1, 1, 0, 0, tzinfo=datetime_mod.timezone.utc)
    """
    if data is None:
        if none_policy == 'return-None':
            return None
        elif none_policy == 'raise':
            raise TimeTypeError('cannot cast None to a datetime')
        else:
            raise KeyError(none_policy)
    elif isinstance(data, str):
        # Canse use ubelt.timeparse(data, default_timezone=default_timezone) here.
        if data == 'now':
            dt = datetime_cls.utcnow()
        else:
            dt = dateutil.parser.parse(data)
    elif isinstance(data, datetime_cls):
        dt = data
    elif isinstance(data, datetime_mod.date):
        dt = dateutil.parser.parse(data.isoformat())
    elif isinstance(data, numbers.Number):
        if math.isnan(data):
            if nan_policy == 'return-None':
                return None
            elif nan_policy == 'raise':
                raise TimeTypeError('cannot cast nan to a datetime')
            else:
                raise KeyError(nan_policy)
        dt = datetime_cls.fromtimestamp(data)
    else:
        raise TimeTypeError('unhandled {}'.format(data))
    dt = ensure_timezone(dt, default=default_timezone)
    return dt


@profile
def coerce_timedelta(delta):
    """
    Parses data that could be associated with a time delta

    Args:
        delta (str | int | float):
            If given as a string, attempt to parse out a time duration.
            Otherwise, interpret pure magnitudes in seconds.

    Returns:
        datetime.timedelta

    TODO:
        move to a util

    Example:
        >>> from watch.utils.util_time import *  # NOQA
        >>> variants = [
        >>>     ['year', 'y'],
        >>>     ['month', 'm', 'mon'],
        >>>     ['day', 'd', 'days'],
        >>>     ['hours', 'hour', 'h'],
        >>>     ['minutes', 'min', 'M'],
        >>>     ['second', 'S', 's', 'secs'],
        >>> ]
        >>> for vs in variants:
        >>>     print('vs = {!r}'.format(vs))
        >>>     ds = []
        >>>     for v in vs:
        >>>         d = coerce_timedelta(f'1{v}')
        >>>         ds.append(d)
        >>>         d = coerce_timedelta(f'1 {v}')
        >>>         ds.append(d)
        >>>     assert ub.allsame(ds)
        >>>     print('ds = {!r}'.format(ds))
        >>> print(coerce_timedelta(10.3))
        >>> print(coerce_timedelta('1y'))
        >>> print(coerce_timedelta('1m'))
        >>> print(coerce_timedelta('1d'))
        >>> print(coerce_timedelta('1H'))
        >>> print(coerce_timedelta('1M'))
        >>> print(coerce_timedelta('1S'))
        >>> print(coerce_timedelta('1year'))
        >>> print(coerce_timedelta('1month'))
        >>> print(coerce_timedelta('1day'))
        >>> print(coerce_timedelta('1hour'))
        >>> print(coerce_timedelta('1min'))
        >>> print(coerce_timedelta('1sec'))
        >>> print(coerce_timedelta('1microsecond'))
        >>> print(coerce_timedelta('1milliseconds'))
        >>> print(coerce_timedelta('1ms'))
        >>> print(coerce_timedelta('1us'))

    References:
        https://docs.python.org/3.4/library/datetime_mod.html#strftime-strptime-behavior
    """
    if isinstance(delta, str):
        try:
            delta = float(delta)
        except ValueError:
            ...

    if isinstance(delta, datetime_mod.timedelta):
        ...
    elif isinstance(delta, numbers.Number):
        try:
            delta = datetime_mod.timedelta(seconds=delta)
        except ValueError:
            if isinstance(delta, float) and math.isnan(delta):
                raise TimeValueError('Cannot coerce nan to a timedelta')
            raise
    elif isinstance(delta, str):

        try:
            ureg = _time_unit_registery()
            seconds = ureg.parse_expression(delta).to('seconds').m
            # timedelta apparently does not have resolution higher than
            # microseconds.
            # https://stackoverflow.com/questions/10611328/strings-ns
            # https://bugs.python.org/issue15443
            delta = datetime_mod.timedelta(seconds=seconds)
        except Exception:
            # Separate the expression into a magnitude and a unit
            import re
            expr_pat = re.compile(
                r'^(?P<magnitude>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
                '(?P<spaces> *)'
                '(?P<unit>.*)$')
            match = expr_pat.match(delta.strip())
            if match:
                parsed = match.groupdict()
                unit = parsed.get('unit', '')
                magnitude = parsed.get('magnitude', '')
            else:
                unit = None
                magnitude = None

            if unit in {'y', 'year', 'years'}:
                delta = datetime_mod.timedelta(days=365 * float(magnitude))
            elif unit in {'d', 'day', 'days'}:
                delta = datetime_mod.timedelta(days=1 * float(magnitude))
            elif unit in {'w', 'week', 'weeks'}:
                delta = datetime_mod.timedelta(days=7 * float(magnitude))
            elif unit == {'m', 'month', 'months'}:
                delta = datetime_mod.timedelta(days=30.437 * float(magnitude))
            elif unit == {'H', 'hour', 'hours'}:
                delta = datetime_mod.timedelta(hours=float(magnitude))
            elif unit == {'M', 'min', 'mins', 'minute', 'minutes'}:
                delta = datetime_mod.timedelta(minutes=float(magnitude))
            elif unit == {'S', 'sec', 'secs', 'second', 'seconds'}:
                delta = datetime_mod.timedelta(seconds=float(magnitude))
            else:
                import pytimeparse  #
                import warnings
                warnings.warn('warning: pytimeparse fallback')
                seconds = pytimeparse.parse(delta)
                if seconds is None:
                    raise Exception(delta)
                delta = datetime_mod.timedelta(seconds=seconds)
                return delta
    else:
        raise TimeTypeError(f'cannot cast {type(delta)} to a timedelta')
    return delta


@profile
def ensure_timezone(dt, default='utc'):
    """
    Gives a datetime_mod a timezone (utc by default) if it doesnt have one

    Arguments:
        dt (datetime.datetime): the datetime to fix
        default (str): the timezone to use if it does not have one.

    Example:
        >>> from watch.utils.util_time import *  # NOQA
        >>> dt = ensure_timezone(datetime_cls.now(), datetime_mod.timezone(datetime_mod.timedelta(hours=+5)))
        >>> print('dt = {!r}'.format(dt))
        >>> dt = ensure_timezone(datetime_cls.utcnow())
        >>> print('dt = {!r}'.format(dt))
        >>> ensure_timezone(datetime_cls.utcnow(), 'utc')
        >>> ensure_timezone(datetime_cls.utcnow(), 'local')
    """
    if dt.tzinfo is not None:
        return dt
    else:
        if isinstance(default, datetime_mod.timezone):
            tzinfo = default
        else:
            if default == 'utc':
                tzinfo = datetime_mod.timezone.utc
            elif default == 'local':
                tzinfo = datetime_mod.timezone(datetime_mod.timedelta(seconds=-time.timezone))
            else:
                raise NotImplementedError
        return dt.replace(tzinfo=tzinfo)


@ub.memoize
def _time_unit_registery():
    import pint
    # Empty registry
    ureg = pint.UnitRegistry(None)
    ureg.define('second = []')
    ureg.define('minute = 60 * second')
    ureg.define('hour = 60 * minute')

    ureg.define('day = 24 * hour')
    ureg.define('month = 30.437 * day')
    ureg.define('year = 365 * day')

    ureg.define('min = minute')
    ureg.define('mon = month')
    ureg.define('sec = second')

    ureg.define('S = second')
    ureg.define('M = minute')
    ureg.define('H = hour')

    ureg.define('d = day')
    ureg.define('m = month')
    ureg.define('y = year')

    ureg.define('s = second')

    ureg.define('millisecond = second / 1000')
    ureg.define('microsecond = second / 1000000')

    ureg.define('ms = millisecond')
    ureg.define('us = microsecond')
    return ureg


_time_unit_registery()


def _format_timedelta(delta, resolution=None):
    """
    TODO format time deltas at some resolution granularity

    Args:
        delta (datetime.timedelta): The timedelta to format

        resolution (Coerceable[datetime.timedelta] | None):
            minimum temporal resolution. If unspecified returns
            an isoformat

    Example:
        >>> from watch.utils.util_time import *  # NOQA
        >>> delta = coerce_timedelta('13months')
        >>> # delta = coerce_timedelta('13months', 'year')

    """
    if resolution is None:
        return str(delta)
    else:
        resolution = coerce_timedelta(resolution)

        # TODO: unit, precision?
        delta / resolution
        raise NotImplementedError
        # s = 13420
        # hours, remainder = divmod(s, 3600)
        # minutes, seconds = divmod(remainder, 60)
        # print('{:02}:{:02}:{:02}'.format(int(hours), int(minutes), int(seconds)))
        # # result: 03:43:40
        # ...



import ubelt as ub


def request_dvc_path(fpath):
    """
    If the file is in a DVC directory and the user has permissions,
    attempt to pull the file from a remote if it doesn't exist locally.
    """
    if not fpath.exists():
        dvc_fpath = fpath.augment(stem=fpath.name, ext='.dvc')
        if dvc_fpath.exists():
            remote = 'aws'  # parametarize
            from watch.utils.simple_dvc import SimpleDVC
            dvc = SimpleDVC()
            dvc.pull(dvc_fpath, remote=remote)
        raise Exception('File {} not exist in a DVC directory'.format(dvc_fpath))


def test_predict_old_fusion_model():
    """
    Check that the old fusion models still work for prediction.

    This tests requires program data as is not exepcted to run on CI in this
    state. We should have a test for our latest-and-greatest models though.
    """
    import watch
    import kwcoco
    import pytest
    # from watch.utils import kwcoco_extensions

    try:
        dvc_dpath = watch.find_smart_dvc_dpath(tags='phase1_data')
    except Exception:
        pytest.skip('dvc path does not exist')

    # model_fpath = dvc_dpath / 'models/fusion/SC-20201117/SC_smt_it_stm_p8_newanns_weighted_raw_v39/SC_smt_it_stm_p8_newanns_weighted_raw_v39_epoch=53-step=2311901.pt'

    # model_fpath = dvc_dpath / 'models/fusion/SC-20201117/SC_smt_it_st_s12_newanns_weighted_rgb_v22/SC_smt_it_st_s12_newanns_weighted_rgb_v22_epoch=117-step=5051933.pt'
    # model_fpath = dvc_dpath / 'models/fusion/SC-20201117/SC_smt_it_stm_p8_newanns_weighted_raw_v39/SC_smt_it_stm_p8_newanns_weighted_raw_v39_epoch=53-step=2311901.pt'
    model_fpath = dvc_dpath / 'models/fusion/eval3_candidates/packages/Drop3_SpotCheck_V323/Drop3_SpotCheck_V323_epoch=18-step=12976.pt'

    from watch.utils.simple_dvc import SimpleDVC
    dvc = SimpleDVC()
    dvc.request(model_fpath)

    # from watch.tasks.fusion import utils
    # method = utils.load_model_from_package(model_fpath)

    # coco_fpath = dvc_dpath / 'Drop1-Aligned-L1-2022-01/data.kwcoco.json'
    coco_fpath = dvc_dpath / 'Aligned-Drop3-TA1-2022-03-10/data_nowv_vali.kwcoco.json'

    if not model_fpath.exists():
        pytest.skip('expected model does not exist')

    if not coco_fpath.exists():
        pytest.skip('expected test data does not exist')

    dset = kwcoco.CocoDataset(coco_fpath)

    output_dpath = ub.Path.appdir('watch/tests/pred/oldmodel').ensuredir()
    pred_fpath = output_dpath / 'pred_bundle/pred.kwcoco.json'

    subset = make_small_kwcoco_subset(dset, output_dpath)

    from watch.tasks.fusion import predict
    pred_kwargs = {
        'test_dataset': subset.fpath,
        'pred_dataset': pred_fpath,
        'package_fpath': model_fpath,
        'chip_overlap': 0.0,
        'gpus': "auto:1",
        'num_workers': 0,
        'set_cover_algo': 'approx',
    }
    kwargs = pred_kwargs  # NOQA
    predict.predict(**pred_kwargs)

    """
    # For devs
    cd $HOME/.cache/watch/tests/pred/oldmodel/
    smartwatch visualize ./test_input.kwcoco.json --viz_dpath=./_viz_test_input
    smartwatch visualize ./pred_bundle/pred.kwcoco.json --viz_dpath=./old_viz_check

    smartwatch intensity_histograms ./pred_bundle/pred.kwcoco.json --dst=./intensity_histo.png --stat probability \
            --valid_range=0:1 --exclude_channels=cirrus

    smartwatch visualize /home/joncrall/data/dvc-repos/smart_watch_dvc/Drop1-Aligned-L1-2022-01/data.kwcoco.json --viz_dpath=./orig_viz_check
    """


def make_small_kwcoco_subset(dset, output_dpath):
    import pytest
    from watch.utils import kwcoco_extensions
    vidid = dset.videos().peek()['id']
    gids = list(dset.images(vidid=vidid))[0:11]
    subset = dset.subset(gids)
    if subset.missing_images(check_aux=True):
        pytest.skip('data has not been pulled down')
    subset.reroot(absolute=True)
    subset.fpath = str(output_dpath / 'test_input.kwcoco.json')
    walker = ub.IndexableWalker(subset.dataset['images'])
    tofix = []
    for path, value in walker:
        # Hack for my sanity
        if path[-1] == 'shear':
            transform_dict = walker[path[:-1]]
            tofix.append(transform_dict)
    import kwimage
    for transform_dict in tofix:
        fixed = kwimage.Affine.coerce(transform_dict).concise()
        transform_dict.clear()
        transform_dict.update(fixed)
    stats = kwcoco_extensions.coco_channel_stats(subset)
    print('stats = {}'.format(ub.urepr(stats, nl=2)))
    subset.dump(subset.fpath, newlines=True)
    return subset


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/tests/test_predict_old_fusion_model.py
    """
    test_predict_old_fusion_model()



import torch
import torch.nn as nn
import torch.nn.functional as F

from watch.tasks.rutgers_material_change_detection.models.base_model import BaseDecoder


class DynamicUNet(BaseDecoder):
    def __init__(self, feat_sizes, in_channels, out_channels, base_feat_channels=64, name="unet"):
        super(DynamicUNet, self).__init__(feat_sizes, out_channels, name)
        self.bfc = base_feat_channels
        self.in_channels = in_channels

        assert type(feat_sizes) is list
        assert len(feat_sizes) == 4

        self.build()

    def build(self):
        # Build input encoder layer.
        self.in_encode_layer = DoubleConv(self.in_channels, self.bfc)

        # Build upsampling layers.

        # last of encoder feats and lowest travel feats
        #  512 + 256 --> 512
        self.up1 = Up(self.feat_sizes[3][0] + self.feat_sizes[2][0], self.feat_sizes[3][0])

        # 1st upsampled feats and 2nd travel feats
        # 256 + 128 --> 256
        self.up2 = Up(self.feat_sizes[3][0] + self.feat_sizes[1][0], self.feat_sizes[2][0])

        # 128 + 64 --> 128
        self.up3 = Up(self.feat_sizes[2][0] + self.feat_sizes[0][0], self.feat_sizes[1][0])

        # 64 + bfc --> 64
        self.up4 = Up(self.feat_sizes[1][0] + self.bfc, self.bfc)

        # Build output convolution layer.
        # bfc --> n_classes
        self.out_conv_layer = OutConv(self.bfc, self.out_channels)

    def forward(self, feats, input):
        """Foward method of Dynamic UNet model.

        Args:
            feats (dict(torch.tensor)): A dict containing keys of form "layerX" and values of tensors of shape [batch_size, channels, height, width].

        Returns:
            (torch.tensor): A tensor of shape [batch_size, out_channels, height, width]
        """
        l0_feats = self.in_encode_layer(input)
        l1_feats = feats["layer1"]
        l2_feats = feats["layer2"]
        l3_feats = feats["layer3"]
        l4_feats = feats["layer4"]

        x = self.up1(l4_feats, l3_feats)
        x = self.up2(x, l2_feats)
        x = self.up3(x, l1_feats)
        x = self.up4(x, l0_feats)

        output = self.out_conv_layer(x)

        return output


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.double_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, bilinear=True):
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2])
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)



import os

import torch
import torchvision

from lydorn_utils import print_utils


def get_backbone(backbone_params):
    set_download_dir()
    if backbone_params["name"] == "unet":
        from torchvision.models.segmentation._utils import _SimpleSegmentationModel
        from frame_field_learning.unet import UNetBackbone

        backbone = UNetBackbone(backbone_params["input_features"], backbone_params["features"])
        backbone = _SimpleSegmentationModel(backbone, classifier=torch.nn.Identity())
    elif backbone_params["name"] == "fcn50":
        backbone = torchvision.models.segmentation.fcn_resnet50(pretrained=backbone_params["pretrained"],
                                                                num_classes=21)
        backbone.classifier = torch.nn.Sequential(*list(backbone.classifier.children())[:-1],
                                                  torch.nn.Conv2d(512, backbone_params["features"], kernel_size=(1, 1),
                                                                  stride=(1, 1)))
    elif backbone_params["name"] == "fcn101":
        backbone = torchvision.models.segmentation.fcn_resnet101(pretrained=backbone_params["pretrained"],
                                                                 num_classes=21)
        backbone.classifier = torch.nn.Sequential(*list(backbone.classifier.children())[:-1],
                                                  torch.nn.Conv2d(512, backbone_params["features"], kernel_size=(1, 1),
                                                                  stride=(1, 1)))

    elif backbone_params["name"] == "deeplab50":
        backbone = torchvision.models.segmentation.deeplabv3_resnet50(pretrained=backbone_params["pretrained"],
                                                                      num_classes=21)
        backbone.classifier = torch.nn.Sequential(*list(backbone.classifier.children())[:-1],
                                                  torch.nn.Conv2d(256, backbone_params["features"], kernel_size=(1, 1),
                                                                  stride=(1, 1)))
    elif backbone_params["name"] == "deeplab101":
        backbone = torchvision.models.segmentation.deeplabv3_resnet101(pretrained=backbone_params["pretrained"],
                                                                       num_classes=21)
        backbone.classifier = torch.nn.Sequential(*list(backbone.classifier.children())[:-1],
                                                  torch.nn.Conv2d(256, backbone_params["features"], kernel_size=(1, 1),
                                                                  stride=(1, 1)))
    elif backbone_params["name"] == "unet_resnet":
        from torchvision.models.segmentation._utils import _SimpleSegmentationModel
        from frame_field_learning.unet_resnet import UNetResNetBackbone

        backbone = UNetResNetBackbone(backbone_params["encoder_depth"], num_filters=backbone_params["num_filters"],
                                      dropout_2d=backbone_params["dropout_2d"],
                                      pretrained=backbone_params["pretrained"],
                                      is_deconv=backbone_params["is_deconv"])
        backbone = _SimpleSegmentationModel(backbone, classifier=torch.nn.Identity())

    elif backbone_params["name"] == "ictnet":
        from torchvision.models.segmentation._utils import _SimpleSegmentationModel
        from frame_field_learning.ictnet import ICTNetBackbone

        backbone = ICTNetBackbone(in_channels=backbone_params["in_channels"],
                                  out_channels=backbone_params["out_channels"],
                                  preset_model=backbone_params["preset_model"],
                                  dropout_2d=backbone_params["dropout_2d"],
                                  efficient=backbone_params["efficient"])
        backbone = _SimpleSegmentationModel(backbone, classifier=torch.nn.Identity())
    else:
        print_utils.print_error("ERROR: config[\"backbone_params\"][\"name\"] = \"{}\" is an unknown backbone!"
                                "If it is a new backbone you want to use, "
                                "add it in backbone.py's get_backbone() function.".format(backbone_params["name"]))
        raise RuntimeError("Specified backbone {} unknown".format(backbone_params["name"]))
    return backbone


def set_download_dir():
    os.environ['TORCH_HOME'] = 'models'  # setting the environment variable



import os
import pickle

import torch
import kwcoco
import hashlib
import numpy as np
from tqdm import tqdm
from einops import rearrange
from tifffile import tifffile
from sklearn.cluster import KMeans

from watch.tasks.rutgers_material_seg_v2.datasets.base_dataset import BaseDataset
from watch.tasks.rutgers_material_seg_v2.utils.util_features import compute_residual_feature


class MaterialPixelDataset(BaseDataset):
    def __init__(self,
                 kwcoco_path,
                 split,
                 image_slice_cfg=None,
                 feature_type='pixel',
                 n_clusters=5,
                 region_shape='square',
                 square_size=5,
                 channels='ALL',
                 seed_num=0,
                 sensor='S2',
                 image_scale_mode='direct',
                 image_norm_mode=None):

        self.channels = channels
        self.n_clusters = n_clusters
        self.square_size = square_size
        self.feature_type = feature_type
        self.region_shape = region_shape
        self.image_norm_mode = image_norm_mode
        self.image_scale_mode = image_scale_mode

        super(MaterialPixelDataset, self).__init__(kwcoco_path,
                                                   split,
                                                   image_scale_mode=image_scale_mode,
                                                   seed_num=seed_num,
                                                   sensor=sensor)

        # Get examples.
        self._build_dataset()

    def get_n_channels(self):
        # TODO: Update channels.
        if self.channels == 'ALL':
            print('WARNING: ALL channels does NOT include cloud channel or materials.')
            self.channels = 'red|green|blue|nir|swir16|swir22'

        if self.feature_type == 'pixel':
            n_channels = len(self.channels.split('|'))
        elif self.feature_type == 'residual':
            n_channels = self.n_clusters
        else:
            raise NotImplementedError

        return n_channels

    def _create_cach_file_name(self):
        input_str = f'{self.split}_{self.feature_type}_{self.n_clusters}_{self.image_scale_mode}_{self.square_size}_{self.sensor}_{self.channels}'
        sha = hashlib.sha256()
        sha.update(input_str.encode())
        output_str = sha.hexdigest()

        return output_str + '.p'

    def _build_dataset(self):
        # Check to see if examples are already generated.
        cache_dir = os.path.join(self.dset_root, '_cache', 'residual_feature')
        os.makedirs(cache_dir, exist_ok=True)
        cache_file_name = self._create_cach_file_name()
        cache_path = os.path.join(cache_dir, cache_file_name)

        if os.path.exists(cache_path) is False:
            # Create examples.
            print(f'No cached examples found at: {cache_path}')

            # Load KWCOCO file.
            self.coco_dset = kwcoco.CocoDataset(self.kwcoco_path)

            # Get all image ids with auxiliary materials.
            self.examples, pixel_data = [], []
            material_class_ids = []
            image_ids = list(self.coco_dset.imgs.keys())
            for image_id in tqdm(image_ids, colour='green', desc='Generating examples'):
                aux_data = self.coco_dset.imgs[image_id]['auxiliary']

                # Check if image has a material mask.
                if aux_data[-1]['channels'] == 'mat_mask':
                    # Load image.
                    image = self._load_norm_image(image_id)
                    image_height, image_width = image.shape[1], image.shape[2]

                    # Load image and get pixels with material labels.
                    file_path = aux_data[-1]['file_name']
                    mat_label_mask = tifffile.imread(file_path)
                    mask_height, mask_width = mat_label_mask.shape[0], mat_label_mask.shape[1]

                    if (image_height != mask_height) or (image_width != mask_height):
                        if (mask_height == image_width) and (mask_width == image_height):
                            mat_label_mask = mat_label_mask[:image_width, :image_height]
                            mat_label_mask = mat_label_mask.T
                        else:
                            mat_label_mask = mat_label_mask[:image_height, :image_width]
                    mask_height, mask_width = mat_label_mask.shape[0], mat_label_mask.shape[1]

                    # Find pixels where there are labels and the pixels contain information.
                    X, Y = np.where((mat_label_mask != 0) & (image.sum(axis=0) != 0))

                    for i in range(X.shape[0]):
                        example = {}
                        x, y = X[i], Y[i]

                        example['center_loc'] = [x, y]
                        example['image_id'] = image_id
                        example['mat_id'] = mat_label_mask[x, y]

                        self.examples.append(example)

                    # Get pixel data for k-Means computation.
                    img_pixels = image[:, X, Y]  # [channels, n_pixels]

                    # Take mean of pixels
                    pixel_data.append(img_pixels)

                    # Get unique material ids.
                    img_mat_ids = list(np.unique(mat_label_mask))
                    material_class_ids.extend(img_mat_ids)

            # Combine pixel data.
            pixel_data = np.concatenate(pixel_data, axis=1)

            # Compute mean and std of this slice of dataset if split == 'train'.
            mean, std = self._compute_dataset_stats(pixel_data)

            # Compute kMeans
            if self.feature_type != 'pixel':
                pixel_data = (pixel_data - mean[:, None]) / (std[:, None] + 1e-5)
                centroid_clusters = self._compute_clusters(pixel_data, self.n_clusters)
            else:
                centroid_clusters = None

            # Save examples and info needed to load from examples.
            cache_object = {
                'examples': self.examples,
                'coco_dset': self.coco_dset,
                'centroid_centers': centroid_clusters,
                'mean': mean,
                'std': std,
                'n_classes': len(set(material_class_ids))
            }
            pickle.dump(cache_object, open(cache_path, 'wb'))
        else:
            # Load cached examples.
            print(f'Loading examples from: {cache_path}')
            cache_object = pickle.load(open(cache_path, 'rb'))

        self.examples = cache_object['examples']
        self.n_classes = cache_object['n_classes']
        self.coco_dset = cache_object['coco_dset']
        self.centroid_clusters = cache_object['centroid_centers']
        self.mean, self.std = cache_object['mean'], cache_object['std']

    def __getitem__(self, index):
        example = self.examples[index]

        # Load pixel or residual feature.
        pixel_data = self._load_pixel_info(self.feature_type, example['image_id'],
                                           example['center_loc'])  # [channels, height, width]

        if self.feature_type == 'pixel':
            pixel_data = pixel_data.reshape([pixel_data.shape[0], -1]).mean(axis=-1)
        else:
            pixel_data = compute_residual_feature(pixel_data,
                                                  self.centroid_clusters,
                                                  img_norm_mode=self.image_norm_mode,
                                                  local_context_mode=f'precomputed|{self.region_shape}')
            if np.isnan(pixel_data).sum() > 0:
                pixel_data = np.nan_to_num(pixel_data)

        output = {}
        output['pixel_data'] = pixel_data
        output['image_id'] = example['image_id']
        output['center_loc'] = example['center_loc']
        output['target'] = example['mat_id'].astype(int)

        return output

    def _load_pixel_info(self, feature_type, image_id, center_loc):
        # Load pixel information.
        if self.region_shape == 'square':
            # Get pixel indices.
            x, y = center_loc
            radius = (self.square_size - 1) // 2
            h0, _ = x - radius, x + radius + 1  # Add 1 because slice is not inclusive
            w0, _ = y - radius, y + radius + 1  # Add 1 because slice is not inclusive

            # Clip box dimensions to not extend past image.
            h0, w0 = np.clip(h0, 0, a_max=None), np.clip(w0, 0, a_max=None)

            # Get pixel values within the square region.
            pixel_data = self._load_norm_image(image_id, crop_slice=[h0, w0, radius * 2 + 1, radius * 2 + 1])
        else:
            raise NotImplementedError

        # TODO: Optionally compute features.

        return pixel_data

    def _compute_clusters(self, pixels, n_clusters):
        """Compute the centroid clusters of the K-Means algorithm.

        Args:
            pixels (np.array): A float array of shape [feature_dim, n_features]
            n_clusters (int): The number of cluster centroids to fit to data.

        Returns:
            np.array: _description_
        """
        print(f'Computing K-Means of {n_clusters} clusters over {pixels.shape[1]} samples ...')
        kmeans = KMeans(n_clusters=n_clusters, random_state=self.seed_num)
        kmeans.fit(pixels.T)
        return kmeans.cluster_centers_.T

    def format_image_data(self, image_data, buffer_mask):
        batch_ssls = rearrange(image_data, 'b c h w -> b c (h w)')
        return {'pixel_data': batch_ssls}


def test_compare_train_valid_dists(args, n_samples=10000):
    # Create train and valid datasets.
    print('Building datasets ...')
    train_dset = MaterialPixelDataset(args.dataset, split='train', feature_type=args.feature_type)
    valid_dset = MaterialPixelDataset(args.dataset, split='valid', feature_type=args.feature_type)

    # Gather some examples from each dataset.
    print(f'\nCollecting {n_samples} samples from each dataset ...')
    ## Train dataset.
    train_indices = np.random.choice(list(range(train_dset.__len__())), n_samples, replace=False)
    train_sample_data = []
    for index in train_indices:
        train_ex = train_dset.__getitem__(index)
        train_sample_data.append(train_ex['pixel_data'])
    train_sample_data = np.stack(train_sample_data, axis=0)

    ## Valid dataset.
    valid_indices = np.random.choice(list(range(valid_dset.__len__())), n_samples, replace=False)
    valid_sample_data = []
    for index in valid_indices:
        valid_ex = valid_dset.__getitem__(index)
        valid_sample_data.append(valid_ex['pixel_data'])
    valid_sample_data = np.stack(valid_sample_data, axis=0)

    print('\nDistribution comparison')
    print('-' * 20)
    print(f'Train dist: {train_sample_data.mean(axis=0)}')
    print(f'Valid dist: {valid_sample_data.mean(axis=0)}')


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset', type=str, default='iarpa_drop4_materials')
    parser.add_argument('--split', type=str, default='valid')
    parser.add_argument('--feature_type', type=str, default='pixel')
    parser.add_argument('--test_type', type=str, default=None)
    parser.add_argument('--seed_num', type=int, default=0)
    args = parser.parse_args()

    # Set random seeds.
    np.random.seed(args.seed_num)
    torch.random.manual_seed(args.seed_num)

    # Run a test.
    if args.test_type is None:
        dataset = MaterialPixelDataset(args.dataset, split=args.split, feature_type=args.feature_type)
        example = dataset.__getitem__(0)
        print(example['pixel_data'].shape)
    elif args.test_type == 'compare_train_valid_dists':
        test_compare_train_valid_dists(args)



from .segmentation_model import segmentation_model
from argparse import Namespace
import pytorch_lightning as pl
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.callbacks import ModelCheckpoint, LearningRateMonitor
from datetime import date


def main(args):
    if type(args) == dict:
        args = Namespace(**args)

    log_dir = '{}/{}/{}/{}/{}'.format(
        args.save_dir,
        args.dataset,
        'multi_image_segment',
        'position_' + str(args.positional_encoding),
        str(date.today()),
        )

    logger = TensorBoardLogger(log_dir)

    model = segmentation_model(hparams=args)

    checkpoint_callback = ModelCheckpoint(monitor='val_epoch_f1', mode='max', save_top_k=1, save_last=True)
    lr_logger = LearningRateMonitor(logging_interval='step')

    trainer = pl.Trainer.from_argparse_args(args,
                                            logger=logger,
                                            callbacks=[checkpoint_callback, lr_logger],
                                            log_every_n_steps=10,
                                            check_val_every_n_epoch=args.check_val_every_n_epoch)
    trainer.fit(model)


if __name__ == '__main__':
    from argparse import ArgumentParser

    parser = ArgumentParser()

    ###train hyperparameters
    parser.add_argument('--max_epochs', type=int, default=50)
    parser.add_argument('--workers', type=int, default=8)
    parser.add_argument('--batch_size', type=int, default=128)
    parser.add_argument('--step_size', type=int, default=10)
    parser.add_argument('--lr_gamma', type=float, default=.1)
    parser.add_argument('--weight_decay', type=float, default=1e-5)
    parser.add_argument('--learning_rate', type=float, default=.0001)
    parser.add_argument('--save_dir', default='watch/tasks/invariants/logs')
    parser.add_argument('--gpus', type=int, default=1)
    parser.add_argument('--drop_rate', type=float, default=.1)

    ###dataset
    parser.add_argument('--dataset', type=str, help='Choose from: spacenet, onera, or kwcoco.', default='kwcoco')

    ### kwcoco arguments
    parser.add_argument('--train_dataset', type=str, default='')
    parser.add_argument('--vali_dataset', type=str, default='')
    parser.add_argument('--sensor', type=str, nargs='+', default=['S2', 'L8'])
    parser.add_argument('--bands', type=str, nargs='+', default=['shared'])

    ### spacenet arguments
    parser.add_argument('--remove_clouds', help='spacenet specific argument', action='store_true')
    parser.add_argument('--normalize_spacenet', help='spacenet specific argument', action='store_true')

    ### onera arguments
    parser.add_argument('--onera_data_folder', help='Path to Onera. Only relevant if train_dataset and/or vali_dataset are onera.', type=str, default='/localdisk0/SCRATCH/watch/onera/')

    #To do: allow for pretrained weights in this architecture
    ### pretraining arguments
    #     parser.add_argument('--pretrained_checkpoint', type=str, help='path to pretrained checkpoint. Leave blank for change detection training without pretraining.', default='')
    #     parser.add_argument('--pretrained_multihead', action='store_true', help='indicate if the pretrained checkpoint was trained in a multihead fashion')
    #     parser.add_argument('--pretrained_encoder_only', action='store_true')

    ### main argument
    parser.add_argument('--patch_size', type=int, default=64)
    parser.add_argument('--num_channels', type=int, default=6)
    parser.add_argument('--pos_class_weight', type=float, help='Weight on positive class for segmentation. Only used on binary labels.', default=1)
    parser.add_argument('--num_images', type=int, default=2)
    parser.add_argument('--attention_layers', type=int, nargs='+', default=[1, 2, 3, 4])
    parser.add_argument('--positional_encoding', action='store_true')
    parser.add_argument('--positional_encoding_mode', type=str, help='addition or concatenation', default='concatenation')
    parser.add_argument('--binary', help='Condense annotations to binary as opposed to site classification. Choose 0 to use classification labels.', type=int, default=1)
    parser.add_argument('--check_val_every_n_epoch', type=int, default=1)
    parser.add_argument('--dataset_style', type=str, default='gridded')
    parser.add_argument('--ignore_boundary', type=int, default=3)
    parser.add_argument('--bas', type=int, default=1)

    args = parser.parse_args()
    main(args)



import torch
import torch.nn as nn

from watch.tasks.rutgers_material_change_detection.models.vit import ViT
from watch.tasks.rutgers_material_change_detection.models.pos_embedding import PositionEncoder

from watch.tasks.rutgers_material_change_detection.utils.util_misc import get_crop_slices


class PatchTransformerEncoder(nn.Module):
    def __init__(
        self,
        task_mode,
        in_channels,
        frame_shape,
        patch_shape,
        max_frames,
        dim=1024,
        n_heads=4,
        n_blocks=6,
        dim_linear_block=2048,
        p_dropout=0.0,
    ):
        super(PatchTransformerEncoder, self).__init__()

        # Check inputs.
        assert type(frame_shape) is int
        assert type(patch_shape) is int

        # Check that patch size matches frame size perfectly.
        assert frame_shape % patch_shape == 0, "Image dimensions must be divisible by the patch size."

        # Generate crop slices.
        self.crop_slices = get_crop_slices(frame_shape, frame_shape, patch_shape, patch_shape, mode="exact")

        # Build model.
        self.dim = dim
        self.n_heads = n_heads
        self.n_blocks = n_blocks
        self.p_dropout = p_dropout
        self.task_mode = task_mode
        self.max_frames = max_frames
        self.in_channels = in_channels
        self.frame_shape = frame_shape
        self.patch_shape = patch_shape
        self.dim_linear_block = dim_linear_block
        self.build()

    def build(self):
        self.encoder = ViT(
            image_size=self.frame_shape,
            patch_size=self.patch_shape,
            num_classes=1,  # Not used so just need a real int here.
            dim=self.dim,
            depth=self.n_blocks,
            heads=self.n_heads,
            mlp_dim=self.dim_linear_block,
            pool="cls",  # just set this to generate an extra token.
            channels=self.in_channels,
            dim_head=64,  # TODO: Not sure what to set for this param.
            dropout=self.p_dropout,
            emb_dropout=self.p_dropout,
        )

    def forward(self, video, datetimes=None):
        """Forward call for PatchTransformer feature encoder.

        Args:
            video (torch.tensor): A tensor of shape [batch_size, n_frames, n_channels, height, width]
            datetimes (list(datetimes.datetimes), optional): A list of datetimes. Used for positional encoding.

        Returns:
            (torch.tensor): TODO
        """

        # Split input video into crops.
        outputs = []
        for crop_slice in self.crop_slices:
            H0, W0, dH, dW = crop_slice

            crop_video = video[:, :, :, H0 : H0 + dH, W0 : W0 + dW]
            # Rearrange crop video to shape [batch_size, n_frames, n_channels * height * width]
            # token_dim = n_channels * height * width
            crop_video = crop_video.flatten(2)

            # Refine patches into tokens of shape [n_frames+1, token_dim]
            crop_tokens = self.encoder.patch_forward(crop_video)

            crop_output = {}
            crop_output["og_height"] = self.frame_shape
            crop_output["og_width"] = self.frame_shape
            crop_output["crop_slice"] = crop_slice
            crop_output["prediction"] = crop_tokens

            outputs.append(crop_output)
        return outputs


class PatchTransformerDecoder(nn.Module):
    def __init__(self, feat_size, task_mode, n_out_channels, decoder_agg_mode, patch_size):
        super(PatchTransformerDecoder, self).__init__()
        self.feat_size = feat_size
        self.task_mode = task_mode
        self.patch_size = patch_size
        self.n_out_channels = n_out_channels
        self.decoder_agg_mode = decoder_agg_mode

        self.build()

    def build(self):
        token_dim = self.feat_size[-1]

        if self.task_mode in ["total_bin_change", "total_sem_change"]:
            out_dim = self.patch_size**2 * self.n_out_channels
            self.decoder = nn.Sequential(nn.LayerNorm(token_dim), nn.Linear(token_dim, out_dim))
        elif self.task_mode in ["ss_splice_change", "ss_splice_change_index", "ss_arrow_of_time"]:
            self.decoder = nn.Sequential(nn.LayerNorm(token_dim), nn.Linear(token_dim, self.n_out_channels))
        else:
            raise NotImplementedError(
                f'Task mode "{self.task_mode}" not implemented for Patch Tranformer Decoder build method.'
            )

    def forward(self, patch_output):
        """[summary]

        Args:
            patch_output (torch.tensor): A tensor of shape [batch_size, n_tokens+1, token_dim].
        Returns:
            (torch.tensor): A torch.tensor of shape [batch_size, n_channels, patch_height, patch_width]
        """

        # Get final token representation.
        if self.decoder_agg_mode == "mean_tokens":
            patch_agg = torch.max(patch_output[:, :-1], dim=1)[0].squeeze(1)  # [batch_size, token_dim]
        elif self.decoder_agg_mode == "max_tokens":
            patch_agg = torch.mean(patch_output[:, :-1], dim=1).squeeze(1)  # [batch_size, token_dim]
        elif self.decoder_agg_mode == "cls_token":
            patch_agg = patch_output[:, -1]
        else:
            raise NotImplementedError(f'Decoder input mode "{self.decoder_agg_mode}" not implemented.')

        pred = self.decoder(patch_agg)

        if self.task_mode in ["total_bin_change", "total_sem_change"]:
            pred = torch.reshape(pred, [-1, self.n_out_channels, self.patch_size, self.patch_size])
        elif self.task_mode in ["ss_splice_change", "ss_splice_change_index", "ss_arrow_of_time"]:
            pass
        else:
            raise NotImplementedError(
                f'Task mode "{self.task_mode}" not implemented for Patch Tranformer Decoder forward method.'
            )

        return pred


class PatchTransformer(nn.Module):
    def __init__(
        self,
        task_mode,
        height,
        width,
        n_channels,
        patch_length,
        max_n_frames,
        out_mode="features",
        token_dim_reduce_factor=2,
        n_heads=4,
        hidden_dim_token_factor=1,
        p_dropout=0.0,
        n_enc_layers=3,
        positional_type="positional",
        patch_seq_normalize=True,
    ):
        super(PatchTransformer, self).__init__()
        self.width = width
        self.height = height
        self.out_mode = out_mode
        self.task_mode = task_mode
        self.n_channels = n_channels
        self.patch_seq_normalize = patch_seq_normalize

        # Check that patches divde equally into image size.
        assert (
            height % patch_length == 0 and width % patch_length == 0
        ), "Image dimensions must be divisible by the patch size."

        # Create crop slices based on image height and width.
        self.crop_slices = get_crop_slices(height, width, patch_length, patch_length, mode="exact")

        # Create linear projection for reducing the dimensionality of input tokens.
        self.patch_feat_length = patch_length**2 * n_channels
        self.token_length = self.patch_feat_length // token_dim_reduce_factor
        # self.project_tokens_layer = nn.Linear(self.patch_feat_length, self.token_length)

        # Build transformer encoder.
        hidden_dim = int(hidden_dim_token_factor * self.token_length)
        encoder_layer = nn.TransformerEncoderLayer(self.patch_feat_length, n_heads, hidden_dim, p_dropout)
        self.transformer = nn.TransformerEncoder(encoder_layer, n_enc_layers)

        # Create positional encoding.
        self.pos_encoder = PositionEncoder(max_n_frames, self.patch_feat_length, pos_type=positional_type)

        # Create output_token
        if self.out_mode == "token_output":
            if task_mode == "total_bin_change":
                self.out_channels = 2
            else:
                raise NotImplementedError(f'Output channels for Patch Transformers task mode "{task_mode}".')

            self.output_token = nn.Embedding(1, self.patch_feat_length)

            # self.channel_comp_layer = nn.Conv2d(n_channels, self.out_channels, kernel_size=3, stride=1, padding=1)
            self.output_resize_layer = nn.Linear(n_channels * patch_length**2, 2 * patch_length**2, bias=False)

    def feature_refinement(self, image):
        # image: [B, F, C, H, W]

        # Create canvas to paste output features.
        B, F, C, H, W = image.shape
        canvas = torch.zeros([B, F, C, H, W], dtype=image.dtype).to(image.device)

        for crop_slice in self.crop_slices:
            # Compute crop of image sequence based on crop parameters.
            h0, w0, h, w = crop_slice
            crop = image[:, :, :, h0 : (h0 + h), w0 : (w0 + w)]

            # Flatten height, width, and channel dimension.
            tokens = crop.flatten(2)

            if self.patch_seq_normalize:
                # Normalize patch.
                tokens = self.normalize_patch(tokens)

            # Linearly project features to lower dimension.
            # tokens = tokens.view(-1, self.patch_feat_length)
            # tokens = self.project_tokens_layer(tokens)
            # tokens = tokens.view(B, -1, self.token_length)

            # Add relative positinal information to tokens.
            # TODO:

            # Add output token if out_mode == 'token'
            if self.out_mode == "token":
                # TODO: Figure this out
                tokens = torch.concat((tokens, self.out_token), dim=0)

            # Pass tokens into transformer encoder.
            tokens = tokens.permute(1, 0, 2)  # [B, F, f] --> [F, B, f]
            output_tokens = self.transformer(tokens)
            output_tokens = output_tokens.permute(1, 0, 2)  # [F, B, f] --> [B, F, f]

            # # TODO: Get output prediction representation.
            # if self.out_mode == 'token':
            #     pass

            # Reshape output prediction.
            out_crop = output_tokens.view(B, F, C, h, w)

            # Put output prediction into canvas
            canvas[:, :, :, h0 : (h0 + h), w0 : (w0 + w)] = out_crop

        return canvas

    def token_output(self, video, active_frames=None):
        # image: [B, F, C, H, W]

        # Create canvas to paste output features.
        B, _, C, H, W = video.shape
        canvas = torch.zeros([B, 2, H, W], dtype=video.dtype).to(video.device)

        for crop_slice in self.crop_slices:
            # Compute crop of video sequence based on crop parameters.
            h0, w0, h, w = crop_slice
            crop = video[:, :, :, h0 : (h0 + h), w0 : (w0 + w)]

            # Flatten height, width, and channel dimension.
            tokens = crop.flatten(2)

            if self.patch_seq_normalize:
                # Normalize patch.
                tokens = self.normalize_patch(tokens, active_frames)

            # Add relative positinal information to tokens.
            tokens = self.pos_encoder(tokens)

            # Add output token if out_mode == 'token'
            output_token = self.output_token(torch.LongTensor([0] * B).to(video.device)).unsqueeze(1)
            tokens = torch.concat((tokens, output_token), dim=1)

            # Pass tokens into transformer encoder.
            tokens = tokens.permute(1, 0, 2)  # [B, F, f] --> [F, B, f]
            output_tokens = self.transformer(tokens)
            output_tokens = output_tokens.permute(1, 0, 2)  # [F, B, f] --> [B, F, f]

            # Reshape output prediction of last ouput token.
            out_crop = output_tokens[:, -1].view(B, C, h, w)

            out_crop = self.output_resize_layer(out_crop.flatten(1)).view(B, 2, h, w)

            # Put output prediction into canvas
            canvas[:, :, h0 : (h0 + h), w0 : (w0 + w)] = out_crop

        # if self.channel_comp_layer:
        #     canvas = self.channel_comp_layer(canvas)

        return canvas

    def encode_image(self, image):
        if self.out_mode == "features":
            return self.image_feature_refinement(image)
        elif self.out_mode == "token_output":
            return self.image_token_output(image)
        else:
            raise NotImplementedError(
                f'Output mode for Patch Transformer encode image not implemented for "{self.out_mode}".'
            )

    def forward(self, data):
        if self.out_mode == "features":
            raise NotImplementedError("Patch Transformer forward: Output mode featues not implemnted.")
            # return self.feature_refinement(data)
        elif self.out_mode == "token_output":
            pred = self.token_output(data["video"], data["active_frames"])
            return {self.task_mode: pred}
        else:
            raise NotImplementedError(
                f'Output mode for Patch Transformer forward not implemented for "{self.out_mode}".'
            )

    def normalize_patch(self, patch, active_frames, eta=1e-3):
        """[summary]

        Note: Only active frames are normalized.

        Args:
            patch (tensor): A tensor of shape [batch_size, n_frames, channels*height*width]

        Returns:
            TODO
        """

        # Get the number of active frames for sequence.
        assert active_frames is not None
        max_frames = (active_frames.sum(dim=1) - 1).int()

        # Compute the mean and std along the number of frames.
        B = patch.shape[0]
        means, stds = [], []
        for b in range(B):
            mean = patch[b, : max_frames[b].item()].mean(dim=1).unsqueeze(1)
            std = patch[b, : max_frames[b].item()].std(dim=1).unsqueeze(1)

            means.append(mean)
            stds.append(std)

        # Convert Means and STDs are on correct device.
        means = torch.tensor(means).unsqueeze(1).unsqueeze(1).to(patch.device)
        stds = torch.tensor(stds).unsqueeze(1).unsqueeze(1).to(patch.device)

        # Normalize the patch.
        patch = (patch - means) / (stds + eta)

        return patch


class PatchTransformer_0(nn.Module):
    def __init__(self, cfg, token_length):
        super().__init__()

        self.mode = cfg.db.mode
        self.token_length = token_length

        n_heads = cfg.db.m.n_heads
        n_enc_layers = cfg.db.m.n_enc_layers
        hidden_dim = cfg.db.m.hidden_dim
        p_dropout = cfg.db.m.p_dropout

        encoder_layer = nn.TransformerEncoderLayer(token_length, n_heads, hidden_dim, p_dropout)
        self.seq_enc = nn.TransformerEncoder(encoder_layer, n_enc_layers)

        # get positional encoder
        if cfg.db.pos_type:
            self.pos_encoder = PositionEncoder(cfg.db.max_seq_len, token_length, cfg.db.pos_type)
        else:
            self.pos_encoder = None

        if self.mode == "change":
            # add change token
            self.change_token = nn.Embedding(1, token_length)

            # add change head
            self.change_head = nn.Sequential(nn.Linear(token_length, 2))
        elif self.mode == "stage":
            self.stage_head = nn.Sequential(nn.Linear(token_length, 14))
        else:
            raise NotImplementedError(self.mode)

    def forward(self, data):
        output = {}

        seq = data["seq"].permute((1, 0, 2))

        # add positional encoding
        if self.pos_encoder:
            seq = self.pos_encoder(seq, data["date_delta"])

        if self.mode == "change":
            # add change token to sequence
            ct = self.change_token(torch.tensor(0).to(seq.device))
            ct = ct.view(1, 1, -1).repeat(1, seq.shape[1], 1)
            seq = torch.cat((seq, ct), dim=0)

            # pass sequence through model
            enc_seq = self.seq_enc(seq)

            output["enc_seq"] = enc_seq[:-1]  # dont need the classification head

            # get change prediction
            output["change_est"] = self.change_head(enc_seq[-1, :]).squeeze()
        elif self.mode == "stage":
            # pass sequence through model
            enc_seq = self.seq_enc(seq)  # [S, B, T]
            S, B, T = enc_seq.shape
            output["enc_seq"] = enc_seq

            stage_est = self.stage_head(enc_seq.permute((2, 0, 1)).flatten(1).T)  # [S*B, C]
            output["stage_est"] = stage_est.view((S, B, 14)).permute((1, 0, 2))

        else:
            raise NotImplementedError(self.mode)

        return output


if __name__ == "__main__":
    frame_size = 40
    n_channels = 4
    patch_size = 20
    n_frames = 5
    encoder = PatchTransformerEncoder("total_bin_mode", n_channels, frame_size, patch_size, n_frames)
    test_input = torch.zeros([1, n_frames, n_channels, frame_size, frame_size])
    output = encoder(test_input)
